"""
Hebrew Phonemizer

Rules implemented:
1. Consonant handling (including special cases)
2. Nikud (vowel) processing
3. Dagesh handling
4. Geresh handling
5. Shva Na prediction
6. Special letter combinations

Reference:
- https://en.wikipedia.org/wiki/Unicode_and_HTML_for_the_Hebrew_alphabet#Compact_table
- https://en.wikipedia.org/wiki/Help:IPA/Hebrew
- https://he.wikipedia.org/wiki/הברה
- https://hebrew-academy.org.il/2020/08/11/איך-הוגים-את-השווא-הנע
- https://hebrew-academy.org.il/2010/03/24/צהרים-נעמי-הגיית-קמץ-לפני-חט
- https://hebrew-academy.org.il/2022/03/03/מלעיל-ומלרע-על-ההטעמה-בעברית
"""

from typing import Literal
from phonikud.variants import Letter
from phonikud import lexicon
import re
from phonikud.utils import sort_stress

SHVA = "\u05b0"
SIN = "\u05c2"
PATAH = "\u05b7"
KAMATZ = "\u05b8"
HATAF_KAMATZ = "\u05b3"
DAGESH = "\u05bc"
HOLAM = "\u05b9"
HIRIK = "\u05b4"
PATAH_LIKE_PATTERN = "[\u05b7-\u05b8]"
KUBUTS = "\u05bb"
TSERE = "\u05b5"
HATAMA = "\u05ab"
VAV_HOLAM = "\u05ba"
DAGESH = "\u05bc"
SEGOL = "\u05b6"


def phonemize_hebrew(
    letters: list[Letter],
    stress_placement: Literal["syllable", "vowel"],
) -> list[str]:
    phonemes = []
    i = 0

    while i < len(letters):
        cur = letters[i]
        prev = letters[i - 1] if i > 0 else None
        next = letters[i + 1] if i < len(letters) - 1 else None
        next_phonemes, skip_offset = letter_to_phonemes(
            cur, prev, next, stress_placement=stress_placement
        )
        # TODO: split into syllables
        # next_letters = next_phonemes, letters[i:i+skip_offset+1]
        phonemes.extend(next_phonemes)
        i += skip_offset + 1

    return phonemes


def letter_to_phonemes(
    cur: Letter,
    prev: Letter | None,
    next: Letter | None,
    stress_placement: Literal["syllable", "vowel"],
) -> tuple[str, int]:
    cur_phonemes = []
    skip_diacritics = False
    skip_consonants = False
    skip_offset = 0

    if lexicon.NIKUD_HASER_DIACRITIC in cur.all_diac:
        skip_consonants = True
        skip_diacritics = True

    elif cur.char == "א" and not cur.diac and prev:
        if next and next.char != "ו":
            skip_consonants = True

    elif (
        cur.char == "י"
        and next
        # Yud without diacritics
        and not cur.diac
        # In middle
        and prev
        # Prev Hirik
        and prev.char + prev.diac != "אֵ"
        # Next Vav has meaning
        and not (next.char == "ו" and next.diac and "\u05b0" not in next.diac)
    ):
        skip_consonants = True

    elif cur.char == "ש" and SIN in cur.diac:
        if (
            next
            and next.char == "ש"
            and not next.diac
            and re.search("[\u05b7\u05b8]", cur.diac)
        ):
            # ^ יששכר
            cur_phonemes.append("sa")
            skip_consonants = True
            skip_diacritics = True
            skip_offset += 1
        else:
            cur_phonemes.append("s")
            skip_consonants = True

    # shin without nikud after sin = sin
    elif cur.char == "ש" and not cur.diac and prev and SIN in prev.diac:
        cur_phonemes.append("s")
        skip_consonants = True

    elif not next and cur.char == "ח" and PATAH in cur.diac:
        # Final Het gnuva
        cur_phonemes.append("ax")
        skip_diacritics = True
        skip_consonants = True

    elif not next and cur.char == "ה" and PATAH in cur.diac:
        # Final He gnuva
        cur_phonemes.append("ah")
        skip_diacritics = True
        skip_consonants = True

    elif not next and cur.char == "ע" and PATAH in cur.diac:
        # Final Ayin gnuva
        cur_phonemes.append("a")
        skip_diacritics = True
        skip_consonants = True

    if cur and "'" in cur.diac and cur.char in lexicon.GERESH_PHONEMES:
        if cur.char == "ת":
            cur_phonemes.append(lexicon.GERESH_PHONEMES.get(cur.char, ""))
            skip_diacritics = True
            skip_consonants = True
        else:
            # Geresh
            cur_phonemes.append(lexicon.GERESH_PHONEMES.get(cur.char, ""))
            skip_consonants = True

    elif DAGESH in cur.diac and cur.char + DAGESH in lexicon.LETTERS_PHONEMES:  # dagesh
        cur_phonemes.append(lexicon.LETTERS_PHONEMES.get(cur.char + DAGESH, ""))
        skip_consonants = True
    elif cur.char == "ו" and lexicon.NIKUD_HASER_DIACRITIC not in cur.all_diac:
        skip_consonants = True

        if prev and "\u05b0" in prev.diac and re.findall("[\u05b9-\u05ba]", cur.diac):
            # ^ לִגְוֹעַ
            cur_phonemes.append("vo")
            skip_diacritics = True
            skip_consonants = True

        elif next and next.char == "ו":
            # One of them has holam

            holams = re.findall("[\u05b9-\u05ba]", cur.diac + next.diac)
            if len(holams) == 2:
                cur_phonemes.append("wo")
                skip_diacritics = True
                skip_offset += 1
            if len(holams) == 1:
                cur_phonemes.append("vo")
                skip_diacritics = True
                skip_offset += 1
            # patah and next.diac empty
            elif cur.diac == next.diac:
                # double Vav
                cur_phonemes.append("vu")
                skip_diacritics = True
                skip_offset += 1
            elif HIRIK in cur.diac:
                cur_phonemes.append("vi")
                skip_diacritics = True
            elif SHVA in cur.diac and not next.diac:
                cur_phonemes.append("v")
                skip_diacritics = True
            elif KAMATZ in cur.diac or PATAH in cur.diac:
                cur_phonemes.append("va")
                skip_diacritics = True
            elif SEGOL in cur.diac:
                cur_phonemes.append("ve")
                skip_diacritics = True
            else:
                # TODO ?
                # skip_consonants = False
                skip_diacritics = False
        else:
            # Single vav

            # Vav with Patah
            if re.search(PATAH_LIKE_PATTERN, cur.diac):
                cur_phonemes.append("va")

            # Tsere
            elif TSERE in cur.diac:
                cur_phonemes.append("ve")
            elif SEGOL in cur.diac:
                cur_phonemes.append("ve")
            # Holam haser
            elif HOLAM in cur.diac:
                cur_phonemes.append("o")
            # Shuruk / Kubutz
            elif KUBUTS in cur.diac or DAGESH in cur.diac:
                cur_phonemes.append("u")
            # Vav with Shva in start
            elif SHVA in cur.diac and not prev:
                cur_phonemes.append("ve")
            # Hirik
            elif HIRIK in cur.diac:
                cur_phonemes.append("vi")
            elif next and not cur.diac:
                # It is fine for now since we use Dicta
                skip_consonants = True
                skip_diacritics = True
            else:
                cur_phonemes.append("v")

            skip_diacritics = True

    if not skip_consonants:
        cur_phonemes.append(lexicon.LETTERS_PHONEMES.get(cur.char, ""))

    if KAMATZ in cur.diac and next and HATAF_KAMATZ in next.diac:
        cur_phonemes.append("o")
        skip_diacritics = True

    nikud_phonemes = []
    if not skip_diacritics:
        nikud_phonemes = [
            lexicon.NIKUD_PHONEMES.get(nikud, "") for nikud in cur.all_diac
        ]
    elif skip_diacritics and lexicon.HATAMA_DIACRITIC in cur.all_diac:
        nikud_phonemes = [lexicon.STRESS_PHONEME]
    cur_phonemes.extend(nikud_phonemes)
    # Ensure the stress is at the beginning of the syllable
    cur_phonemes = sort_stress(cur_phonemes, stress_placement)
    cur_phonemes = [
        p for p in cur_phonemes if all(i in lexicon.SET_PHONEMES for i in p)
    ]
    # Remove empty phonemes
    cur_phonemes = [p for p in cur_phonemes if p]
    return cur_phonemes, skip_offset



import torch
from torch import nn
from tqdm import tqdm, trange
from torch.utils.tensorboard import SummaryWriter

from torch.utils.data import DataLoader
from config import TrainArgs
from transformers.models.bert.tokenization_bert_fast import BertTokenizerFast
from evaluate import evaluate_model
from phonikud.src.model.phonikud_model import PhoNikudModel


def train_model(
    model: PhoNikudModel,
    tokenizer: BertTokenizerFast,
    train_dataloader: DataLoader,
    val_dataloader: DataLoader,
    args: TrainArgs,
    writer: SummaryWriter,
):
    model.train()
    optimizer = torch.optim.AdamW(model.parameters(), lr=args.learning_rate)
    criterion = nn.BCEWithLogitsLoss()
    scheduler = torch.optim.lr_scheduler.ConstantLR(optimizer, total_iters=args.epochs)
    # Boosted on GPU
    scaler = torch.amp.GradScaler(args.device, enabled="cuda" in args.device)

    step = args.pre_training_step
    best_val_score = float("inf")
    early_stop_counter = 0

    for epoch in trange(args.epochs, desc="Epoch"):
        pbar = tqdm(
            enumerate(train_dataloader), desc="Train iter", total=len(train_dataloader)
        )
        for _, (inputs, targets) in pbar:
            optimizer.zero_grad()

            # Log learning rate
            for param_group in optimizer.param_groups:
                lr = param_group["lr"]
                writer.add_scalar("LR", lr, step)

            inputs = inputs.to(args.device)
            targets = targets.to(args.device)
            # ^ shape: (batch_size, n_chars_padded, n_active_components)
            output = model(inputs)
            # ^ shape: (batch_size, n_chars_padded, 4)

            # Get only the logits for the components we're training on
            active_logits = output.additional_logits[
                :, 1:-1
            ]  # skip BOS and EOS symbols

            loss = criterion(active_logits, targets.float())

            scaler.scale(loss).backward()
            # Unscale gradients before clipping
            scaler.unscale_(optimizer)

            # Clip gradients
            torch.nn.utils.clip_grad_norm_(model.parameters(), max_norm=1.0)

            scaler.step(optimizer)
            scaler.update()
            scheduler.step()

            pbar.set_description(f"Train iter (L={loss.item():.4f})")
            step += 1

            # Log total loss
            writer.add_scalar("Loss/train", loss.item(), step)

            # Always save "last"
            if args.checkpoint_interval and step % args.checkpoint_interval == 0:
                last_dir = f"{args.output_dir}/last"
                print(f"💾 Saving last checkpoint at step {step} to: {last_dir}")
                model.save_pretrained(last_dir)
                tokenizer.save_pretrained(last_dir)

            # Val
            if args.checkpoint_interval and step % args.checkpoint_interval == 0:
                # Evaluate and maybe save "best"
                val_score = evaluate_model(model, val_dataloader, args, writer, step)

                if val_score < best_val_score:
                    best_val_score = val_score
                    best_dir = f"{args.output_dir}/best"
                    print(
                        f"🏆 New best model at step {step} (val_score={val_score:.4f}), saving to: {best_dir}"
                    )
                    model.save_pretrained(best_dir)
                    tokenizer.save_pretrained(best_dir)
                    early_stop_counter = 0
                else:
                    print(
                        f"📉 No improvement at step {step} (no_improvement_counter={early_stop_counter})"
                    )
                    early_stop_counter += 1

                if (
                    args.early_stopping_patience
                    and early_stop_counter > args.early_stopping_patience
                ):
                    break

        # Break batch loop
        if (
            args.early_stopping_patience
            and early_stop_counter >= args.early_stopping_patience
        ):
            print(
                f"🚨 Early stopping at epoch {epoch}, step {step}. No improvement in validation score for {args.early_stopping_patience} steps."
            )
            break

        # Evaluate each epoch
        # evaluate_model(model, val_dataloader, args, writer, step)

    final_dir = f"{args.output_dir}/loss_{loss.item():.2f}"
    print(f"🚀 Saving trained model to: {final_dir}")
    model.save_pretrained(final_dir)
    tokenizer.save_pretrained(final_dir)



"""
fuzzing with millions of words to make sure it works fine
git clone https://github.com/thewh1teagle/hebrew_diacritized
uv run examples/fuzz.py hebrew_diacritized/data
"""

import sys
from pathlib import Path
from tqdm import tqdm
from phonikud import phonemize

target_dir = sys.argv[1]
files = sorted(Path(target_dir).glob("**/*.txt"), key=lambda f: f.stat().st_size)

total_size_bytes = sum(f.stat().st_size for f in files)
total_size_mb = total_size_bytes / (1024 * 1024)
print(f"Found {len(files)} files with total size of {total_size_mb:.2f} MB")

for file in tqdm(files, desc="Files"):
    # First pass: count lines without storing them
    with open(file, "r", encoding="utf-8") as f:
        num_lines = sum(1 for _ in f)

    # Second pass: process lines with progress bar
    with open(file, "r", encoding="utf-8") as f:
        for line in tqdm(f, total=num_lines, desc=f"{file.name}", leave=False):
            phonemes = phonemize(line)
            # print(phonemes)



from phonikud import phonemize

text = """
sfd;kljasdf;lkjasdflkjsdf;lfjaksf'fajl;k23409740278423@#$@!4@!#%#^$#$%&$%^*%^*%$^=156465165 5156s1fd 321. as0f0 as321f3as21dfsaf 
asdfsafsafsd fasf asf dכגדכש דגכ ש
כד שכגשד כשד כדג 
dsf fd/j.lkj/kj['j;jpoiup9iu987755]
我喜欢学习新技术和挑战自己。
أنا أحب تعلم التقنيات الجديدة وتحدي نفسي.
Esdf$fsdADsf

"""


def test_invalid_input():
    print(phonemize(text))


test_invalid_input()



from phonikud import lexicon
from phonikud.variants import Letter
from .expander import Expander
from phonikud.utils import (
    get_letters,
    normalize,
    post_normalize,
    post_clean,
    add_milra_hatama,
    mark_shva_na,
    sort_hatama,
)
from typing import Callable, Literal
import regex as re
from phonikud.hebrew import phonemize_hebrew


class Phonemizer:
    # TODO: is that enough? what if there's punctuation around? other chars?
    fallback_pattern = r"[a-zA-Z]+"

    def __init__(self):
        self.expander = Expander()

    def phonemize(
        self,
        text: str,
        preserve_punctuation: bool,
        preserve_stress: bool,
        use_expander: bool,
        use_post_normalize: bool,  # For TTS
        predict_stress: bool,
        predict_shva_nah: bool,
        stress_placement: Literal["syllable", "vowel"],
        schema: Literal["plain", "modern"],
        fallback: Callable[[str], str] = None,
    ) -> str | list[str]:
        # normalize
        text = normalize(text)

        def fallback_replace_callback(match: re.Match):
            word = match.group(0)

            if self.expander.dictionary.dict.get(word):
                # skip
                # TODO: better API
                return word
            phonemes = fallback(word).strip()
            # TODO: check that it has only IPA?!
            for c in phonemes:
                lexicon.ADDITIONAL_PHONEMES.add(c)
            return phonemes

        if fallback is not None:
            text = re.sub(self.fallback_pattern, fallback_replace_callback, text)

        if use_expander:
            text = self.expander.expand_text(text)

        def heb_replace_callback(match: re.Match, original_text: str):
            word = match.group(0)
            start_offset = match.start()
            if start_offset > 0 and original_text[start_offset - 1] == "[":
                # Skip if it starts with [ as it's used for hyper phonemes
                return word

            if predict_shva_nah:
                mark_shva_na(word)
            if lexicon.HATAMA_DIACRITIC not in word and predict_stress:
                word = add_milra_hatama(word)
            letters: list[Letter] = get_letters(word)
            letters = sort_hatama(letters)

            phonemes: list[str] = phonemize_hebrew(
                letters,
                stress_placement=stress_placement,
            )
            phonemes = "".join(phonemes)
            # syllables = get_syllables(phonemes)

            # phonemes_text = "".join(phonemes)
            # # if predict_stress and lexicon.STRESS not in phonemes_text and syllables:
            # #     if len(syllables) == 1:
            # #         syllables[-1] = lexicon.STRESS + syllables[-1]
            # #         syllables[-1] = "".join(sort_stress(syllables[-1]))
            # #     elif any(
            # #         remove_nikud(word).endswith(i) for i in lexicon.MILHEL_PATTERNS
            # #     ) or phonemes_text.endswith("ax"):
            # #         # insert lexicon.STRESS in the first character of syllables[-2]
            # #         syllables[-2] = lexicon.STRESS + syllables[-2]
            # #         syllables[-2] = "".join(sort_stress(syllables[-2]))
            # #     else:
            # #         # insert in syllables[-1]
            # #         syllables[-1] = lexicon.STRESS + syllables[-1]
            # #         syllables[-1] = "".join(sort_stress(syllables[-1]))

            # phonemes = "".join(syllables)
            if use_post_normalize:
                phonemes = post_normalize(phonemes)

            if schema == "modern":
                # We'll keep this feature simple for now
                for k, v in lexicon.MODERN_SCHEMA.items():
                    phonemes = re.sub(k, v, phonemes)
            return phonemes

        text = re.sub(
            lexicon.HE_PATTERN, lambda match: heb_replace_callback(match, text), text
        )

        def hyper_phonemes_callback(match: re.Match):
            """
            Expand hyper phonemes into normal phonemes
            eg. [hello](/hɛˈloʊ/) -> hɛˈloʊ
            """
            matched_phonemes = match.group(2)
            for c in matched_phonemes:
                lexicon.ADDITIONAL_PHONEMES.add(c)
            return matched_phonemes  # The phoneme is in the second group

        text = re.sub(r"\[(.+?)\]\(\/(.+?)\/\)", hyper_phonemes_callback, text)

        if not preserve_punctuation:
            text = "".join(i for i in text if i not in lexicon.PUNCTUATION or i == " ")
        if not preserve_stress:
            text = "".join(i for i in text if i not in [lexicon.STRESS_PHONEME])
        if use_post_normalize:
            # We don't keep hypens in the output, but we should replace it with space
            text = post_clean(text)
        return text



import num2words
from .number_names import NUMBER_NAMES
import re


def add_diacritics(words: str):
    new_words = []
    for word in words.split():
        if NUMBER_NAMES.get(word):
            new_words.append(NUMBER_NAMES[word])
        elif NUMBER_NAMES.get(word[1:]):
            # With Vav or Bet
            new_words.append(NUMBER_NAMES[word[0]] + NUMBER_NAMES[word[1:]])
        else:
            new_words.append(word)
    return " ".join(new_words)


def num_to_word(maybe_number: str) -> str:
    def replace_number(match):
        num: str = match.group()
        suffix, prefix = "", ""
        # prefix
        if not num.startswith("-") and not num[0].isdigit():
            prefix = num[0]
            num = num[1:]
        if not num[-1].isdigit():
            suffix = num[-1]
            num = num[:-1]
        words = num2words.num2words(num, lang="he", ordinal=False)
        words_with_diacritics = add_diacritics(words)
        return (
            f"{prefix.strip()} {words_with_diacritics.strip()} {suffix.strip()}".strip()
        )

    # Replace all whole numbers in the string
    result = re.sub(r"[^\d\-]?-?\d+(?:[\.,]\d+)?[^\d]?", replace_number, maybe_number)

    return result



from phonikud import lexicon, phonemize
from pathlib import Path
import pandas as pd
import warnings
from phonikud.lexicon import STRESS_PHONEME

TEST_STRESS = True


def test_phonemize_hebrew_sanity():
    with_stress = phonemize("שָׁ֫לוֹם", preserve_stress=True, schema="plain")
    without_stress = phonemize("שָׁ֫לוֹם", preserve_stress=False, schema="plain")
    assert without_stress == "ʃalom" and lexicon.STRESS_PHONEME in with_stress


def run_phoneme_check(
    nikkud, expected_ipa, using_stress, *, filename, line_number, warn_only
):
    output = phonemize(
        nikkud,
        preserve_stress=using_stress,
        use_post_normalize=True,
        use_expander=True,
        predict_shva_nah=False,
        schema="plain",
    )
    ref = expected_ipa
    if not using_stress:
        ref = ref.replace(STRESS_PHONEME, "")
        output = output.replace(STRESS_PHONEME, "")

    msg = f"Incorrect phonemization: {ref} != {output} ({nikkud}; {'with' if using_stress else 'no'} stress) FILE {filename} LINE {line_number}"

    if warn_only and ref != output:
        warnings.warn(msg)
    elif not warn_only:
        assert ref == output, msg


def check_file(filename: str, warn_only=False):
    df = pd.read_csv(filename)
    for line_number, row in enumerate(df.itertuples()):
        nikkud_forms = [row.hebrew_with_nikkud]
        # handle Hebrew punctuation variants
        nikkud_forms += (
            [
                row.hebrew_with_nikkud.replace('"', "״"),
                row.hebrew_with_nikkud.replace("'", "׳"),
            ]
            if any(c in row.hebrew_with_nikkud for c in ['"', "'"])
            else []
        )

        for nikkud in nikkud_forms:
            run_phoneme_check(
                nikkud,
                row.ipa,
                using_stress=TEST_STRESS and STRESS_PHONEME in row.ipa,
                filename=filename,
                line_number=line_number,
                warn_only=warn_only,
            )


def test_phonemize_hebrew_manual():
    base = Path(__file__).parent / "phonemize_test_tables"
    check_file(base / "basic.csv")
    check_file(base / "dictionary.csv")
    check_file(base / "advanced.csv", warn_only=True)



"""
Examples:
    Evaluate on a specific file:
        uv run src/run/check_error_rate.py -m ./ckpt/last -f ./data/train/example.txt -n 100

    Evaluate on training split:
        uv run src/run/check_error_rate.py -m ./ckpt/last --split train -n 100

    Evaluate on eval/dev split:
        uv run src/run/check_error_rate.py -m ./ckpt/last --split eval -n 100
"""

import random
from pathlib import Path
from tqdm import tqdm
from jiwer import wer, cer
from transformers import AutoTokenizer
from src.model.phonikud_model import (
    PhoNikudModel,
    remove_nikud,
    PHONETIC_NIKUD,
    NIKUD_HASER,
)
from src.train.config import BASE_PATH, TrainArgs
from phonikud.utils import normalize
from tap import Tap
from src.train.utils import read_lines


class Args(Tap):
    file: str = None
    split: str = None  # 'train' or 'eval'
    model: str = BASE_PATH / "./ckpt/best"
    device: str = "cuda"
    n: int = 100

    def configure(self):
        self.add_argument("--file", "-f", help="Evaluate on specific file")
        self.add_argument(
            "--split", choices=["train", "eval"], help="Evaluate on train/eval split"
        )
        self.add_argument("--model", "-m")
        self.add_argument("--n", "-n", type=int)
        return super().configure()


def main():
    args = Args().parse_args()

    if args.file:
        lines = Path(args.file).read_text(encoding="utf-8").splitlines()
    elif args.split:
        train_lines, eval_lines = read_lines(
            TrainArgs.data_dir,
            val_split=TrainArgs.val_split,
            split_seed=TrainArgs.split_seed,
        )
        lines = train_lines if args.split == "train" else eval_lines
    else:
        raise ValueError("You must provide either --file or --split")

    lines = [normalize(l.strip()) for l in lines if l.strip()]
    lines = random.sample(lines, min(args.n, len(lines)))

    model = PhoNikudModel.from_pretrained(args.model, trust_remote_code=True)
    tokenizer = AutoTokenizer.from_pretrained(args.model)
    model.to(args.device)
    model.eval()

    gts, preds = [], []

    for line in tqdm(lines, desc="Evaluating"):
        src = remove_nikud(line, additional=PHONETIC_NIKUD)
        if not src:
            continue
        pred = model.predict([src], tokenizer, mark_matres_lectionis=NIKUD_HASER)[0]
        gts.append(remove_nikud(line))
        preds.append(remove_nikud(normalize(pred)))

    w = wer(gts, preds)  # wer
    c = cer(gts, preds)  # cer

    print("\n📊 Evaluation Report:")
    print(f"📝 Lines evaluated: {len(gts)}")

    print(f"📉 WER: {w:.3f} (Acc: {(1 - w) * 100:.1f}%)")
    print(f"🔡 CER: {c:.3f} (Acc: {(1 - c) * 100:.1f}%)")


if __name__ == "__main__":
    main()



"""
uv run src/test.py --device cuda
"""

from src.model.phonikud_model import (
    PhoNikudModel,
    NIKUD_HASER,
    remove_nikud,
    PHONETIC_NIKUD,
)
from src.train.config import BASE_PATH
from transformers import AutoTokenizer
from transformers.models.bert.tokenization_bert_fast import BertTokenizerFast
from tap import Tap
from phonikud.utils import normalize


class RunArgs(Tap):
    model: str = BASE_PATH / "./ckpt/best"  # --model, -m
    device = "cuda"
    file: str = BASE_PATH / "./data/eval/dummy.txt"

    def configure(self):
        self.add_argument("--model", "-m", help="Path to the model checkpoint")
        return super().configure()


def main():
    args = RunArgs().parse_args()
    model = PhoNikudModel.from_pretrained(args.model, trust_remote_code=True)
    tokenizer: BertTokenizerFast = AutoTokenizer.from_pretrained(args.model)
    model.to(args.device)
    model.eval()

    with open(args.file, "r", encoding="utf-8") as fp:
        for src in fp:
            src = normalize(src.strip())
            without_nikud = remove_nikud(src, additional=PHONETIC_NIKUD)
            if not without_nikud:
                continue
            predicted = model.predict(
                [without_nikud], tokenizer, mark_matres_lectionis=NIKUD_HASER
            )[0]
            predicted = normalize(predicted)
            src, predicted = remove_nikud(src), remove_nikud(predicted)
            print()
            print(src == predicted)
            print(without_nikud)
            print(src)
            print(predicted)


if __name__ == "__main__":
    main()



"""
uv pip install phonemizer-fork espeakng-loader
uv run examples/with_fallback.py
"""

from phonemizer.backend.espeak.wrapper import EspeakWrapper
import phonemizer
import espeakng_loader
import phonikud

EspeakWrapper.set_library(espeakng_loader.get_library_path())
EspeakWrapper.set_data_path(espeakng_loader.get_data_path())


def fallback(text: str) -> str:
    """
    Make sure to return valid IPA phonemes and punctuation here
    """
    return phonemizer.phonemize(text)


text = """
מֵרְאִיָּה מְמֻחְשֶׁבֶת עַד Large Language Models: הֵרָשְׁמוּ לְמַחֲזוֹר הֶחָדָשׁ שֶׁל קוּרְס פִּתּוּחַ AI וְ-Deep Learning
"""

phonemes = phonikud.phonemize(
    text, preserve_punctuation=True, preserve_stress=True, fallback=fallback
)
print(phonemes)



"""
https://en.wikipedia.org/wiki/Unicode_and_HTML_for_the_Hebrew_alphabet#Compact_table

TODO: add to phonikud?
"""

import regex as re
import phonikud

VOWEL_DIACS = [chr(i) for i in range(0x05B1, 0x05BC)] + [chr(0x05C7)] + [chr(0x5BD)]

STRESS = "\u05ab"
SHVA = "\u05b0"
DAGESH = "\u05bc"


def sort_diacritics(word: str):
    def sort_diacritics_callback(match):
        letter = match.group(1)
        diacritics = "".join(sorted(match.group(2)))  # Sort diacritics
        return letter + diacritics

    return re.sub(r"(\p{L})(\p{M}+)", sort_diacritics_callback, word)


def has_vowel_diacs(s: str):
    if s == "וּ":
        return True
    return any(i in s for i in VOWEL_DIACS)


def get_syllables(word: str) -> list[str]:
    letters = phonikud.utils.get_letters(word)
    syllables, cur = [], ""
    vowel_state = False

    i = 0
    while i < len(letters):
        letter = letters[i]
        has_vowel = has_vowel_diacs(str(letter)) or (i == 0 and SHVA in letter.all_diac)
        # Look ahead
        vav1 = i + 2 < len(letters) and letters[i + 2].char == "ו"
        vav2 = i + 3 < len(letters) and letters[i + 3].char == "ו"

        if has_vowel:
            if vowel_state:
                syllables.append(cur)
                cur = str(letter)
            else:
                cur += str(letter)
            vowel_state = True
        else:
            cur += str(letter)

        i += 1

        # If two וs are coming: force current syllable to end, and join both וs as next syllable
        if vav1 and vav2:
            if cur:
                # Finish current syllable
                syllables.append(cur + str(letters[i]))
                cur = ""
            cur = str(letters[i + 1]) + str(letters[i + 2])
            i += 3  # skip past the double-vav
            vowel_state = True

        # If one ו is coming, end the syllable now
        elif vav1 and letters[i + 1].diac:
            if cur:
                syllables.append(cur)
                cur = ""
            vowel_state = False

    if cur:
        syllables.append(cur)
    # print(syllables)
    return syllables


def add_stress_to_syllable(s: str):
    letters = phonikud.utils.get_letters(s)
    letters[0].all_diac = STRESS + letters[0].all_diac
    return "".join(letter.char + letter.all_diac for letter in letters)


def add_stress(word: str, syllable_position: int):
    syllables: list[str] = get_syllables(word)

    if not syllables:
        return word  # no syllables, return original word

    # Normalize negative indices
    if syllable_position < 0:
        syllable_position += len(syllables)

    # Clamp to valid range
    syllable_position = max(0, min(syllable_position, len(syllables) - 1))

    stressed_syllable = syllables[syllable_position]
    stressed_syllable = add_stress_to_syllable(stressed_syllable)
    syllables[syllable_position] = stressed_syllable

    return "".join(syllables)



"""
wget https://huggingface.co/thewh1teagle/phonikud/resolve/main/phonikud-1.0.int8.onnx
uv sync
uv run python examples/file.py <some file>
"""

from phonikud_onnx import Phonikud
from phonikud import lexicon
import sys


def main():
    phonikud = Phonikud("./phonikud-1.0.int8.onnx")
    file = sys.argv[1]
    with open(file) as fp:
        for line in fp:
            with_diacritics = phonikud.add_diacritics(
                line, mark_matres_lectionis=lexicon.NIKUD_HASER_DIACRITIC
            )
            print(with_diacritics.strip())


if __name__ == "__main__":
    main()



"""
uv run  examples/hyper_phonemes.py
"""

from phonikud import phonemize

text = """
שָׁל֫וֹם לְכֻלָּ֫ם! הַיּ֫וֹם נִלְמַ֫ד אַנְגְּלִ֫ית:
[hello](/hɛˈloʊ/) [world](/wɜːrld/)
[אנציקלופדיה](/ʔantsikloˈpedja/)
"""

phonemes = phonemize(text)
print(phonemes)



from pathlib import Path
from typing import List, Tuple

import humanize
import torch
from phonikud.src.model.phonikud_model import (
    NIKUD_HASER,
    remove_nikud,
)
from tqdm import tqdm


def print_model_size(model):
    def count_params(module):
        return sum(p.numel() for p in module.parameters())

    def pretty(n):
        return humanize.intword(n)

    print("🔍 Model breakdown:")
    print(f"  ⚙️  MLP: {pretty(count_params(model.mlp))} parameters")
    print(f"  📘 Menaked: {pretty(count_params(model.menaked))} parameters")
    print(f"  🧠  BERT: {pretty(count_params(model.bert))} parameters")


def read_lines(
    data_path: str,
    max_context_length: int = 2048,
    val_split: float = 0.1,
    split_seed: int = 42,
) -> Tuple[List[str], List[str]]:
    files = list(Path(data_path).glob("**/*.txt"))
    total_bytes = sum(f.stat().st_size for f in files)

    lines = []
    with tqdm(
        total=total_bytes, desc="📚 Loading text files...", unit="B", unit_scale=True
    ) as pbar:
        for file in files:
            with open(file, "r", encoding="utf-8") as fp:
                for line in fp:
                    pbar.update(len(line.encode("utf-8")))
                    # Split lines into chunks if they are too long
                    while len(line) > max_context_length:
                        lines.append(line[:max_context_length].strip())
                        line = line[max_context_length:]

                    if line.strip():
                        lines.append(line.strip())

    # Preprocess lines (remove nikud and other components)
    lines = [remove_nikud(i, additional=NIKUD_HASER) for i in lines]

    # Split into train and validation sets
    split_idx = int(len(lines) * (1 - val_split))
    torch.manual_seed(split_seed)
    idx = torch.randperm(len(lines))
    train_lines = [lines[i] for i in idx[:split_idx]]
    val_lines = [lines[i] for i in idx[split_idx:]]

    # Print samples
    print("🛤️ Train samples:")
    for i in train_lines[:3]:
        print(f"\t{i}")
    print("🧪 Validation samples:")
    for i in val_lines[:3]:
        print(f"\t{i}")

    return train_lines, val_lines



from .model import OnnxModel
import re


class Phonikud:
    def __init__(self, model_path: str):
        self.model = OnnxModel(model_path)

    def add_diacritics(
        self, sentences: list | str, mark_matres_lectionis: str | None = None
    ) -> str:
        """
        Adds nikud (Hebrew diacritics) to the given text.

        Parameters:
        - sentences (list | str): A string or a list of strings to be processed. Each string should not exceed 2048 characters.
        - mark_matres_lectionis (str | None, optional): A string used to mark nikud male. For example, if set to '|',
            "לִימּוּדָיו" will be returned as "לִי|מּוּדָיו". Default is None (no marking).

        Returns:
        - str: The text with added diacritics.
        """

        if isinstance(sentences, str):
            sentences = [sentences]
        result = self.model.predict(
            sentences, mark_matres_lectionis=mark_matres_lectionis
        )
        return result[0]

    def get_nikud_male(self, text: str, mark_matres_lectionis: str):
        """
        Based on given mark character remove the mark character to keep it as nikud male
        """
        return text.replace(mark_matres_lectionis, "")

    def get_nikud_haser(self, text: str):
        """
        Based on given mark_matres_lectionis remove the nikud nikud male character along with the mark character
        """
        return re.sub(r".\|", "", text)  # Remove {char}{matres_lectionis}



import onnxruntime as ort
import numpy as np
from tokenizers import Tokenizer
import re

# Constants
NIKUD_CLASSES = [
    "",
    "<MAT_LECT>",
    "\u05bc",
    "\u05b0",
    "\u05b1",
    "\u05b2",
    "\u05b3",
    "\u05b4",
    "\u05b5",
    "\u05b6",
    "\u05b7",
    "\u05b8",
    "\u05b9",
    "\u05ba",
    "\u05bb",
    "\u05bc\u05b0",
    "\u05bc\u05b1",
    "\u05bc\u05b2",
    "\u05bc\u05b3",
    "\u05bc\u05b4",
    "\u05bc\u05b5",
    "\u05bc\u05b6",
    "\u05bc\u05b7",
    "\u05bc\u05b8",
    "\u05bc\u05b9",
    "\u05bc\u05ba",
    "\u05bc\u05bb",
    "\u05c7",
    "\u05bc\u05c7",
]
SHIN_CLASSES = ["\u05c1", "\u05c2"]  # shin, sin
MAT_LECT_TOKEN = "<MAT_LECT>"
MATRES_LETTERS = list("אוי")
ALEF_ORD = ord("א")
TAF_ORD = ord("ת")
STRESS_CHAR = "\u05ab"  # "ole" symbol marks stress
MOBILE_SHVA_CHAR = "\u05bd"  # "meteg" symbol marks shva na (mobile shva)
PREFIX_CHAR = "|"


def is_hebrew_letter(char):
    return ALEF_ORD <= ord(char) <= TAF_ORD


def is_matres_letter(char):
    return char in MATRES_LETTERS


nikud_pattern = re.compile(r"[\u05B0-\u05BD\u05C1\u05C2\u05C7]")


def remove_nikkud(text):
    return nikud_pattern.sub("", text)


class OnnxModel:
    def __init__(
        self, model_path, tokenizer_name="dicta-il/dictabert-large-char-menaked"
    ):
        # Load the tokenizer
        self.tokenizer = Tokenizer.from_pretrained(tokenizer_name)

        # Create ONNX Runtime session
        self.session = ort.InferenceSession(model_path)
        self.input_names = [input.name for input in self.session.get_inputs()]
        self.output_names = [output.name for output in self.session.get_outputs()]

    def _create_inputs(self, sentences: list[str], padding: str):
        # Tokenize inputs using tokenizers library
        encodings = []
        for sentence in sentences:
            encoding = self.tokenizer.encode(sentence)
            encodings.append(encoding)

        # Get the max length for padding
        max_len = max(len(enc.ids) for enc in encodings) if padding == "longest" else 0

        # Prepare batch inputs
        input_ids = []
        attention_mask = []
        offset_mapping = []

        for encoding in encodings:
            ids = encoding.ids
            masks = [1] * len(ids)
            offsets = encoding.offsets

            # Pad if needed
            if padding == "longest" and len(ids) < max_len:
                padding_length = max_len - len(ids)
                ids = ids + [self.tokenizer.token_to_id("[PAD]")] * padding_length
                masks = masks + [0] * padding_length
                offsets = offsets + [(0, 0)] * padding_length

            input_ids.append(ids)
            attention_mask.append(masks)
            offset_mapping.append(offsets)

        # Convert to numpy arrays for ONNX Runtime
        return {
            "input_ids": np.array(input_ids, dtype=np.int64),
            "attention_mask": np.array(attention_mask, dtype=np.int64),
            # Token type IDs might be needed depending on your model
            "token_type_ids": np.zeros_like(np.array(input_ids, dtype=np.int64)),
        }, offset_mapping

    def predict(self, sentences, mark_matres_lectionis=None, padding="longest"):
        sentences = [remove_nikkud(sentence) for sentence in sentences]
        inputs, offset_mapping = self._create_inputs(sentences, padding)

        # Run inference
        outputs = self.session.run(self.output_names, inputs)

        # Process outputs based on output names
        nikud_idx = self.output_names.index("nikud_logits")
        shin_idx = self.output_names.index("shin_logits")
        nikud_logits = outputs[nikud_idx]
        shin_logits = outputs[shin_idx]

        additional_idx = self.output_names.index("additional_logits")
        additional_logits = outputs[additional_idx]

        # Get predictions
        nikud_predictions = np.argmax(nikud_logits, axis=-1)
        shin_predictions = np.argmax(shin_logits, axis=-1)
        stress_predictions = (additional_logits[..., 1] > 1).astype(np.int32)
        mobile_shva_predictions = (additional_logits[..., 2] > 1).astype(np.int32)
        prefix_predictions = (additional_logits[..., 3] > 1).astype(np.int32)

        ret = []
        for sent_idx, (sentence, sent_offsets) in enumerate(
            zip(sentences, offset_mapping)
        ):
            # Assign the nikud to each letter
            output = []
            prev_index = 0
            for idx, offsets in enumerate(sent_offsets):
                # Add anything we missed
                if offsets[0] > prev_index:
                    output.append(sentence[prev_index : offsets[0]])
                if offsets[1] - offsets[0] != 1:
                    continue

                # Get next char
                char = sentence[offsets[0] : offsets[1]]
                prev_index = offsets[1]
                if not is_hebrew_letter(char):
                    output.append(char)
                    continue

                nikud = NIKUD_CLASSES[nikud_predictions[sent_idx][idx]]
                shin = (
                    "" if char != "ש" else SHIN_CLASSES[shin_predictions[sent_idx][idx]]
                )

                # Check for matres lectionis
                if nikud == MAT_LECT_TOKEN:
                    if not is_matres_letter(char):
                        nikud = ""  # Don't allow matres on irrelevant letters
                    elif mark_matres_lectionis is not None:
                        nikud = mark_matres_lectionis
                    else:
                        output.append(char)
                        continue

                stress = (
                    STRESS_CHAR
                    if stress_predictions is not None
                    and stress_predictions[sent_idx][idx] == 1
                    else ""
                )
                mobile_shva = (
                    MOBILE_SHVA_CHAR
                    if mobile_shva_predictions is not None
                    and mobile_shva_predictions[sent_idx][idx] == 1
                    else ""
                )

                prefix = (
                    PREFIX_CHAR
                    if prefix_predictions is not None
                    and prefix_predictions[sent_idx][idx] == 1
                    else ""
                )

                output.append(char + shin + nikud + stress + mobile_shva + prefix)
            output.append(sentence[prev_index:])
            ret.append("".join(output))

        return ret



from pathlib import Path


def test_has_no_breakpoint():
    files = Path(__file__).parent.joinpath("../phonikud").resolve().glob("**/*.py")
    for file in files:
        print(file)

        with open(file, "r") as f:
            content = f.readlines()
        for line_number, line in enumerate(content, 1):
            if "breakpoint()" in line:
                print(f"Error in {file}:{line_number} - 'breakpoint()' found")
                assert False, f"Error in {file}:{line_number} - 'breakpoint()' found"



"""
High level phonemize functions
"""

from .phonemize import Phonemizer
from .utils import normalize  # noqa: F401
from typing import Callable, Literal

phonemizer = Phonemizer()


def phonemize(
    text: str,
    preserve_punctuation=True,
    preserve_stress=True,
    use_expander=True,
    use_post_normalize=True,  # For TTS
    predict_stress=True,
    predict_shva_nah=True,
    stress_placement: Literal["syllable", "vowel"] = "vowel",
    schema: Literal["plain", "modern"] = "modern",
    fallback: Callable[[str], str] = None,
) -> str:
    """
    Set stress_at_start=True to place stress at syllable start.
    """
    phonemes = phonemizer.phonemize(
        text,
        preserve_punctuation=preserve_punctuation,
        preserve_stress=preserve_stress,
        fallback=fallback,
        use_expander=use_expander,
        use_post_normalize=use_post_normalize,
        predict_stress=predict_stress,
        schema=schema,
        predict_shva_nah=predict_shva_nah,
        stress_placement=stress_placement,
    )
    return phonemes



"""
Dictionaries are tab separated key value words
"""

from pathlib import Path
import json
import re
from phonikud.utils import remove_nikud
from phonikud.utils import normalize
from phonikud import lexicon
import unicodedata

files = Path(__file__).parent.joinpath("../data").glob("*.json")
# Sort in reverse order to prioritize the most recent and best
order = {"bronze": 1, "silver": 2, "gold": 3}
files = sorted(
    files, key=lambda f: order.get(next((x for x in order if x in f.stem), ""), 0)
)


class Dictionary:
    def __init__(self):
        self.dict = {}
        self.load_dictionaries()

    def load_dictionaries(self):
        for file in files:
            with open(file, "r", encoding="utf-8") as f:
                dictionary: dict = json.load(f)
                normalized_dictionary = {}

                # normalize nikud keys
                for k, v in dictionary.items():
                    k = normalize(k)
                    # Ensure not empty
                    if k and v:
                        normalized_dictionary[k] = v
                self.dict.update(normalized_dictionary)

    def replace_hebrew_only_callback(self, match: re.Match[str]) -> str:
        source: str = match.group(0)
        # decomposite
        source = unicodedata.normalize("NFD", source)
        raw_lookup = self.dict.get(source)

        without_nikud_lookup = self.dict.get(remove_nikud(source))
        with_nikud_lookup = self.dict.get(normalize(source))
        # Compare without nikud ONLY if source has no nikud
        if raw_lookup:
            return raw_lookup
        if without_nikud_lookup:
            return without_nikud_lookup
        elif with_nikud_lookup:
            return with_nikud_lookup
        return source

    def replace_non_whitespace_callback(self, match: re.Match[str]) -> str:
        raw_source: str = match.group(0)
        if raw_source.isnumeric():
            return raw_source

        raw_lookup = self.dict.get(raw_source)

        # Compare without nikud ONLY if source has no nikud
        if raw_lookup:
            return raw_lookup
        # search by only ', space, regular nikud, alphabet
        raw_source = re.sub(
            lexicon.HE_PATTERN, self.replace_hebrew_only_callback, raw_source
        )
        return raw_source

    def expand_text(self, text: str) -> str:
        """
        TODO: if key doesn't have diacritics expand even diacritized words
        """
        text = re.sub(r"\S+", self.replace_non_whitespace_callback, text)

        return text



"""
uv sync --extra onnx
uv run export.py --model ../phonikud/ckpt/last
"""

import torch
import onnx
from pathlib import Path
from onnxruntime.quantization import quantize_dynamic, QuantType
from argparse import ArgumentParser
import sys

sys.path.append(str(Path(__file__).parent / "../phonikud/src"))
from model.phonikud_model import (
    PhoNikudModel,
)  # TODO: add it as package for autocomplete


def parse_args():
    parser = ArgumentParser(description="Export and quantize model")
    parser.add_argument(
        "--model",
        type=str,
        default="dicta-il/dictabert-large-char-menaked",
        help="Name of the model to export and quantize.",
    )
    return parser.parse_args()


args = parse_args()

# Config
dynamic_axes = True
dynamic_axes_dict = None
batch_size = 1
sequence_length = 128

int8_model_path = "phonikud-1.0.int8.onnx"
fp32_model_path = "phonikud-1.0.onnx"


# Define a model wrapper class for ONNX export
class ModelForONNX(torch.nn.Module):
    def __init__(self, model):
        super().__init__()
        self.model = model

    def forward(self, input_ids, attention_mask, token_type_ids):
        # Create a dictionary of inputs as expected by the model's forward method
        inputs = {
            "input_ids": input_ids,
            "attention_mask": attention_mask,
            "token_type_ids": token_type_ids,
        }

        # Call the model with the inputs dictionary
        outputs = self.model(inputs)

        # Extract the three outputs we want to export
        return (outputs.nikud_logits, outputs.shin_logits, outputs.additional_logits)


def main():
    # Load your custom model
    print(f"Loading model from: {args.model}")

    # Option 1: If your model can be loaded directly from checkpoint
    model = PhoNikudModel.from_pretrained(args.model)

    # Option 2: If you need to load from a saved state dict
    # model = PhoNikudModel(config)
    # state_dict = torch.load(os.path.join(args.model, "pytorch_model.bin"))
    # model.load_state_dict(state_dict)

    model.eval()

    # Wrap the model for ONNX export
    wrapped_model = ModelForONNX(model)

    # Create dummy inputs
    dummy_input_ids = torch.ones((batch_size, sequence_length), dtype=torch.long)
    dummy_attention_mask = torch.ones((batch_size, sequence_length), dtype=torch.long)
    dummy_token_type_ids = torch.zeros((batch_size, sequence_length), dtype=torch.long)

    # Define dynamic axes if requested
    if dynamic_axes:
        dynamic_axes_dict = {
            "input_ids": {0: "batch_size", 1: "sequence_length"},
            "attention_mask": {0: "batch_size", 1: "sequence_length"},
            "token_type_ids": {0: "batch_size", 1: "sequence_length"},
            "nikud_logits": {0: "batch_size", 1: "sequence_length"},
            "shin_logits": {0: "batch_size", 1: "sequence_length"},
            "additional_logits": {0: "batch_size", 1: "sequence_length"},
        }

    # Create model folder
    Path(fp32_model_path).parent.mkdir(exist_ok=True)

    print(f"Exporting onnx model to: {fp32_model_path}...")
    # Export with the wrapped model
    torch.onnx.export(
        wrapped_model,
        args=(dummy_input_ids, dummy_attention_mask, dummy_token_type_ids),
        f=fp32_model_path,
        input_names=["input_ids", "attention_mask", "token_type_ids"],
        output_names=["nikud_logits", "shin_logits", "additional_logits"],
        dynamic_axes=dynamic_axes_dict,
        opset_version=14,
        do_constant_folding=True,
        export_params=True,
        verbose=False,
    )
    print("✅ ONNX model export completed!")

    # Verify the exported model
    print("Verifying ONNX model integrity...")
    onnx_model = onnx.load(fp32_model_path)
    onnx.checker.check_model(onnx_model)
    print("🎉 ONNX model verification successful! Ready to use.")

    # Perform dynamic quantization (INT8)
    print(f"Quantizing model to INT8: {int8_model_path}...")
    quantize_dynamic(
        model_input=fp32_model_path,
        model_output=int8_model_path,
        weight_type=QuantType.QInt8,  # Use QuantType.QUInt8 for unsigned weights if needed
    )
    print("✅ INT8 quantized model export completed!")


if __name__ == "__main__":
    main()



"""
Copied from: https://huggingface.co/dicta-il/dictabert-large-char-menaked/blob/main/BertForDiacritization.py
"""

from dataclasses import dataclass
from typing import List, Optional, Tuple, Union
import torch
from torch import nn
from transformers.utils import ModelOutput
from transformers import BertPreTrainedModel, BertModel, BertTokenizerFast
import re

# MAT_LECT => Matres Lectionis, known in Hebrew as Em Kriaa.
MAT_LECT_TOKEN = "<MAT_LECT>"
NIKUD_CLASSES = [
    "",
    MAT_LECT_TOKEN,
    "\u05bc",
    "\u05b0",
    "\u05b1",
    "\u05b2",
    "\u05b3",
    "\u05b4",
    "\u05b5",
    "\u05b6",
    "\u05b7",
    "\u05b8",
    "\u05b9",
    "\u05ba",
    "\u05bb",
    "\u05bc\u05b0",
    "\u05bc\u05b1",
    "\u05bc\u05b2",
    "\u05bc\u05b3",
    "\u05bc\u05b4",
    "\u05bc\u05b5",
    "\u05bc\u05b6",
    "\u05bc\u05b7",
    "\u05bc\u05b8",
    "\u05bc\u05b9",
    "\u05bc\u05ba",
    "\u05bc\u05bb",
    "\u05c7",
    "\u05bc\u05c7",
]
SHIN_CLASSES = ["\u05c1", "\u05c2"]  # shin, sin


@dataclass
class MenakedLogitsOutput(ModelOutput):
    nikud_logits: torch.FloatTensor = None
    shin_logits: torch.FloatTensor = None

    def detach(self):
        return MenakedLogitsOutput(
            self.nikud_logits.detach(), self.shin_logits.detach()
        )


@dataclass
class MenakedOutput(ModelOutput):
    loss: Optional[torch.FloatTensor] = None
    logits: Optional[MenakedLogitsOutput] = None
    hidden_states: Optional[Tuple[torch.FloatTensor]] = None
    attentions: Optional[Tuple[torch.FloatTensor]] = None


@dataclass
class MenakedLabels(ModelOutput):
    nikud_labels: Optional[torch.FloatTensor] = None
    shin_labels: Optional[torch.FloatTensor] = None

    def detach(self):
        return MenakedLabels(self.nikud_labels.detach(), self.shin_labels.detach())

    def to(self, device):
        return MenakedLabels(self.nikud_labels.to(device), self.shin_labels.to(device))


class BertMenakedHead(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.config = config

        if not hasattr(config, "nikud_classes"):
            config.nikud_classes = NIKUD_CLASSES
            config.shin_classes = SHIN_CLASSES
            config.mat_lect_token = MAT_LECT_TOKEN

        self.num_nikud_classes = len(config.nikud_classes)
        self.num_shin_classes = len(config.shin_classes)

        # create our classifiers
        self.nikud_cls = nn.Linear(config.hidden_size, self.num_nikud_classes)
        self.shin_cls = nn.Linear(config.hidden_size, self.num_shin_classes)

    def forward(
        self, hidden_states: torch.Tensor, labels: Optional[MenakedLabels] = None
    ):
        # run each of the classifiers on the transformed output
        nikud_logits = self.nikud_cls(hidden_states)
        shin_logits = self.shin_cls(hidden_states)

        loss = None
        if labels is not None:
            loss_fct = nn.CrossEntropyLoss()
            loss = loss_fct(
                nikud_logits.view(-1, self.num_nikud_classes),
                labels.nikud_labels.view(-1),
            )
            loss += loss_fct(
                shin_logits.view(-1, self.num_shin_classes), labels.shin_labels.view(-1)
            )

        return loss, MenakedLogitsOutput(nikud_logits, shin_logits)


class BertForDiacritization(BertPreTrainedModel):
    def __init__(self, config):
        super().__init__(config)
        self.config = config
        self.bert = BertModel(config, add_pooling_layer=False)

        classifier_dropout = (
            config.classifier_dropout
            if config.classifier_dropout is not None
            else config.hidden_dropout_prob
        )
        self.dropout = nn.Dropout(classifier_dropout)

        self.menaked = BertMenakedHead(config)

        # Initialize weights and apply final processing
        self.post_init()

    def forward(
        self,
        input_ids: Optional[torch.Tensor] = None,
        attention_mask: Optional[torch.Tensor] = None,
        token_type_ids: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.Tensor] = None,
        head_mask: Optional[torch.Tensor] = None,
        inputs_embeds: Optional[torch.Tensor] = None,
        labels: Optional[torch.Tensor] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ) -> Union[Tuple[torch.Tensor], MenakedOutput]:
        return_dict = (
            return_dict if return_dict is not None else self.config.use_return_dict
        )

        bert_outputs = self.bert(
            input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )

        hidden_states = bert_outputs[0]
        hidden_states = self.dropout(hidden_states)

        loss, logits = self.menaked(hidden_states, labels)

        if not return_dict:
            return (loss, logits) + bert_outputs[2:]

        return MenakedOutput(
            loss=loss,
            logits=logits,
            hidden_states=bert_outputs.hidden_states,
            attentions=bert_outputs.attentions,
        )

    def predict(
        self,
        sentences: List[str],
        tokenizer: BertTokenizerFast,
        mark_matres_lectionis: str = None,
        padding="longest",
    ):
        sentences = [remove_nikkud(sentence) for sentence in sentences]
        # assert the lengths aren't out of range
        assert all(
            len(sentence) + 2 <= tokenizer.model_max_length for sentence in sentences
        ), (
            f"All sentences must be <= {tokenizer.model_max_length}, please segment and try again"
        )

        # tokenize the inputs and convert them to relevant device
        inputs = tokenizer(
            sentences,
            padding=padding,
            truncation=True,
            return_tensors="pt",
            return_offsets_mapping=True,
        )
        offset_mapping = inputs.pop("offset_mapping")
        inputs = {k: v.to(self.device) for k, v in inputs.items()}

        # calculate the predictions
        logits = self.forward(**inputs, return_dict=True).logits
        nikud_predictions = logits.nikud_logits.argmax(dim=-1).tolist()
        shin_predictions = logits.shin_logits.argmax(dim=-1).tolist()

        ret = []
        for sent_idx, (sentence, sent_offsets) in enumerate(
            zip(sentences, offset_mapping)
        ):
            # assign the nikud to each letter!
            output = []
            prev_index = 0
            for idx, offsets in enumerate(sent_offsets):
                # add in anything we missed
                if offsets[0] > prev_index:
                    output.append(sentence[prev_index : offsets[0]])
                if offsets[1] - offsets[0] != 1:
                    continue

                # get our next char
                char = sentence[offsets[0] : offsets[1]]
                prev_index = offsets[1]
                if not is_hebrew_letter(char):
                    output.append(char)
                    continue

                nikud = self.config.nikud_classes[nikud_predictions[sent_idx][idx]]
                shin = (
                    ""
                    if char != "ש"
                    else self.config.shin_classes[shin_predictions[sent_idx][idx]]
                )

                # check for matres lectionis
                if nikud == self.config.mat_lect_token:
                    if not is_matres_letter(char):
                        nikud = ""  # don't allow matres on irrelevant letters
                    elif mark_matres_lectionis is not None:
                        nikud = mark_matres_lectionis
                    else:
                        continue

                output.append(char + shin + nikud)
            output.append(sentence[prev_index:])
            ret.append("".join(output))

        return ret


ALEF_ORD = ord("א")
TAF_ORD = ord("ת")


def is_hebrew_letter(char):
    return ALEF_ORD <= ord(char) <= TAF_ORD


MATRES_LETTERS = list("אוי")


def is_matres_letter(char):
    return char in MATRES_LETTERS


nikud_pattern = re.compile(r"[\u05B0-\u05BD\u05C1\u05C2\u05C7]")


def remove_nikkud(text):
    return nikud_pattern.sub("", text)



"""
Add stress (Hat'ama) diacritic using Phonikud
"""

from phonikud.syllables import add_stress

data = {"אוֹמֶרֶת": -2, "דֶרֶךְ": -2,
        "גִּימִיקִים": -3, "קוֹרֶה": -1, "רַדְיוֹ": -2}
data = {"רֵפוֹרְמוֹת": -2}
for k, v in data.items():
    stressed = add_stress(k, v)
    print(stressed)



"""
Quick Nakdan https://nakdanlive.dicta.org.il
Pro Nakdan https://nakdanpro.dicta.org.il
Manual Nakdan https://www.yo-yoo.co.il/tools/nikud

uv run examples/phonemize.py
"""

from phonikud import phonemize

sentences = [
    # "הִיא אָמְרָה לִי",  # Hirik with Yod vowel
    # "מָה קוֹרֶה?",  # Vav with Holam vowel
    # "הוּא קֵרֵחַ",  # Het gnuva
    # "גִ'ירָפָה זַ'רְגּוֹן",  # Geresh
    # "שָׁלוֹם שׂוֹרֵר",  # Shin Sin
    # "לְבִיבָה יְלָדִים שְׁאֵלָה בְּהֵמָה זְעָקוֹת בְּרֹאשׁ גּוֹרְרִים מְצַפְצְפִים מִרְפְּאוֹתֵיהֶם",  # Shva na
    # "בְּרוֹשׁ סְבִיבָה בְּרָגִים שְׁפֵלָה בְּרֵכָה זְנָבוֹת גּוֹרְמִים הָלַכְתְּ וּבְרָכָה",  # Shva nax
    # "אַתָּה חַיָּב לִי 50 שֶׁקֶל וְהַסֵּמֶל הוּא ₪",  # numbers, symbols
    # "כל",  # dictionary
    # "אֶלֶגַנְטִי",  # lookup
    # "כָּל",  # match to one without nikud
    # "יִשְׂרָאֵל הוּקְמָה בִּשְׁנַת 1948",
    # "₪",
    # "מָחָר בְּשָׁעָה 22:00",  # time to word
    # "אוֹרוֹת ווֹלְטֶר ווֹלֶנְטַיְן אֻרְווֹת אוֹרְחִים אָרְבוּ הוֹוֶה וּוּבְּרִים וָו וִי וּוִיקְטוֹר",  # Vavs
    # "לְמַיִם אֱלֹהֵי",
    # "שְׁמַע אָח שֶׁלִּי, תִּרְאֶה. נִרְאֶה לִי שֶׁזֶּה עוֹבֵד!",
    # "גִ'ירָפָה",
    # "יִהְיֶה",
    # "בְּמָה",
    # "בְּיִשְׂרָאֵל.",
    # "אֲוִויר",
    # "לַכִּיווּן",
    # """קִיוּוּ""",
    # "אֻרְווֹת",
    # "וּוִיקְטוֹר",
    # "אֲוִויר",
    # "מְאוּוְרֶרֶת",
    # "טוֹוִים",
    # "מִגְוַון",
    # "הַקַּווְקָז",
    # "בְּסִיוֹוֹן",
    # "צוֹוַחַת",
    # "צוֹוְחִים",
    # "וּוֶנְדֶּר",
    # "וּוַעֲדַת",
    # "שָׁלוֹם",
    # "וּמֶה",
    # "הַתְּרוּמָה",
    # "English",
    # "בַּרִ֫יא",
    # "זֶ֫ה הָיָ֫ה קוֹרֵ֫עַ מִצְּח֫וֹק הֵ֫ם קוֹרְעִ֫ם אוֹתִ֫י",
    # "וּבֽתוֹלְד֫וֹת",
    # "וִירוּשָׁלַיִם",
    # "וְאַתָּה",
    # "אוּרִי",
    # "אַוָּרִי",
    # "וָואלָה",
    # "ווֹלְטֶר",
    # "אוֹרִי",
    # "אֻרִי",
    # "גִּ׳ירָהּ",
    # "גִּ'רַ֫פָה",
    # "זָ'קֵ֫ט",
    # "תַּ'מְבִין",
    # "כֹּחַ",
    # "מֹחַ",
    # "כָּל",
    # "רֵיחַ",
    # "אֶלֶף עֶרֶב",
    # "לִוְיָתָ֫ן",
    # "יִשָּׂשכָר",
    # "יִשָּׂשכָ֫ר",
    # "יַאלְלָה",
    # "אִיָּ֫ר",
    # "קוֹרֶה",
    # "אָח",
    # "שְׁמוּרִים",
    # "עֶ֫רֶב",
    # "מַגְנִיבָה",
    # "אוֹפַנַּיִים",
    # "אֵיתָנוּת",
    # "רֵיחַ",
    # "אֵיתָנוּת",
    # "אֵיר֫וֹפָּה",
    # "צָהֳרַ֫יִם",
    # "דַּוָָר",
    # "עַכְשָׁ֫יו",
    # "כׇּל אֶחָד יָכוֹל!",
    # "הַמָּלֵא",
    # "רוּחַ",
    # "אֶנְצִיקְלוֹפֶּ֫דְיָה",
    # 'צַהַ"ל',
    # "שָׁלוֹם",
    # "אֶנְצִיקְלוֹפֶּדְיָה",
    # "אוֹפַנַּייִם @"c
    # "קֵרֵחַ",
    # "שָׁלוֹם חֲבֵרִים! מָה שְׁלוֹמְכֶם?",
    # "אֶנְצִיקְלוֹפֶּ֫דְיָה",
    # "וַאלְלָה",
    # "בָּךְ",
    # "חֲתֻונָּה",
    # "דִּוּ֫וּחַ",
    # "שֶׁמְּֽשַׁוֶּ֫קֶת",
    # "וּמֽק֫וֹם",
    # "ו֫וֹלְטֶר",  # double vav start
    # "הֲו֫וֹלְטֶר",  # double vav middle
    # "וַואטְסְאַפּ"
    # "עַל-כֵּן",
    # "לְהַגְבִּיהַּ",
    # "לִשְׁקֹעַ",
    # "כׇּל",
    # "אֵיבָרָיו",
    # "סְתָיו",
    # "לֽאַט",
    # "רֽאִי",
    # "אֵלָיו",
    # "ו֫וֹלְטֶר",
    # "צַהַ״ל",
    # 'צַהַ"ל',
    # "לִרְאוֹת",
    # "סִי֯֫יַּמְתִּי",
    # "טְרַא֯מְפּ"
    "אֲוְורִירִי",
    "לִוְויְיתָן",
    "טַוָוס",
]
# print(get_syllables("אֵלָיו"))
for sentence in sentences:
    phonemes = phonemize(
        sentence,
        preserve_punctuation=True,
        use_post_normalize=True,
        use_expander=True,
        predict_stress=True,
        predict_shva_nah=True,
        stress_placement="vowel",
        schema="modern",
    )
    print(phonemes)



from torch import nn
import torch
from config import TrainArgs
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm


def evaluate_model(
    model,
    val_dataloader: DataLoader,
    args: TrainArgs,
    writer: SummaryWriter,
    step,
):
    model.eval()  # Set the model to evaluation mode
    val_loss = 0
    criterion = nn.BCEWithLogitsLoss()

    with torch.no_grad():  # No gradients needed during evaluation
        for inputs, targets in tqdm(val_dataloader, desc="Evaluating 🧠"):
            inputs = inputs.to(args.device)
            targets = targets.to(args.device)

            output = model(inputs)
            active_logits = output.additional_logits[:, 1:-1]  # skip BOS and EOS

            loss = criterion(active_logits, targets.float())
            val_loss += loss.item()

    val_loss /= len(val_dataloader)  # Average over all validation batches
    writer.add_scalar("Loss/val", val_loss, step)  # Log validation loss
    print(f"✅ Validation Loss after step {step}: {val_loss:.4f} 📉")
    return val_loss



from datetime import datetime
from .numbers import num_to_word

# Mapping of month names in Hebrew with diacritics (Gregorian months)
MONTHS = {
    1: "יָ֫נוּאָר",
    2: "פֶ֫בְרוּאָר",
    3: "מֵ֫רְץ",
    4: "אֵפְרִיל",
    5: "מַאי",
    6: "י֫וּנִי",
    7: "י֫וּלִי",
    8: "א֫וֹגֻסְט",
    9: "סֶפְּטֶ֫מְבֶּר",
    10: "אוֹקְט֫וֹבֶּר",
    11: "נוֹבֶ֫מְבֶּר",
    12: "דֶּצֶ֫מְבֶּר",
}

# Mapping of day names in Hebrew with diacritics
DAYS = {
    0: "יוֹם רִאשׁוֹן",
    1: "יוֹם שֵׁנִי",
    2: "יוֹם שְׁלִישִׁי",
    3: "יוֹם רֵבִיעִי",
    4: "יוֹם חֲמִישִׁי",
    5: "יוֹם שִׁישִׁי",
    6: "יוֹם שַׁבָּת",
}


def date_to_word(word: str, include_day_name=False) -> str:
    """
    Converts a given date string in formats (YYYY-MM-DD, YYYY.MM.DD, YYYY/MM/DD) to Hebrew date format with diacritics.
    Returns the original word if it's not a valid date.
    """
    separators = ["-", ".", "/"]
    orders = [("%Y", "%m", "%d"), ("%d", "%m", "%Y")]
    date_formats = [sep.join(order) for order in orders for sep in separators]

    for date_format in date_formats:
        try:
            # Try parsing the word with each date format
            date_obj = datetime.strptime(word, date_format)

            # Get the Hebrew day name with diacritics
            day_name = DAYS[date_obj.weekday()]

            # Convert month to Hebrew name with diacritics
            month_name = MONTHS[date_obj.month]
            day = num_to_word(str(date_obj.day))
            year = num_to_word(str(date_obj.year))

            text = f"{day} בֵּ{month_name} {year}"
            if include_day_name:
                text = f"{day_name}, {text}"
            return text
        except ValueError:
            continue
    return word



"""
See https://github.com/savoirfairelinux/num2words/blob/master/num2words/lang_HE.py
"""

# TODO: add nikud hints

ZERO = {"אפס": "אֶ֫פֶס"}


ONES = {
    "אחת": "אַחַת",
    "אחד": "אֶחָד",
    "ראשונה": "רִאשׁוֹנָה",
    "ראשון": "רִאשׁוֹן",
    "ראשונות": "רִאשׁוֹנוֹת",
    "ראשונים": "רִאשׁוֹנִים",
    "שתיים": "שְׁתַּ֫יִם",
    "שניים": "שְׁנַ֫יִם",
    "שתי": "שְׁתֵּי",
    "שני": "שְׁנֵי",
    "שנייה": "שְׁנִיָּה",
    "שניות": "שְׁנִיּוֹת",
    "שלוש": "שָׁלוֹשׁ",
    "שלושה": "שְׁלוֹשָׁה",
    "שלושת": "שְׁל֫וֹשֶׁת",
    "שלישית": "שְׁלִישִׁית",
    "שלישי": "שְׁלִישִׁי",
    "שלישיות": "שְׁלִישִׁיּוֹת",
    "שלישיים": "שְׁלִישִׁיִּים",
    "ארבע": "אַ֫רְבַּע",
    "ארבעה": "אַרְבַּעָה",
    "ארבעת": "אַרְבַּ֫עַת",
    "רביעית": "רֵבִיעִית",
    "רביעי": "רֵבִיעִי",
    "רביעיות": "רֵבִיעִיוֹת",
    "רביעיים": "רֵבִיעִיִּים",
    "חמש": "חָמֵשׁ",
    "חמישה": "חֲמִשָּׁה",
    "חמשת": "חֲמֵ֫שֶׁת",
    "חמישית": "חֲמִישִּׁית",
    "חמישי": "חֲמִישִּׁי",
    "חמישיות": "חֲמִישִּׁיוֹת",
    "חמישיים": "חֲמִישִּׁיִּים",
    "שש": "שֵׁשׁ",
    "שישה": "שִׁשָּׁה",
    "ששת": "שֵׁ֫שֶׁת",
    "שישית": "שִׁשִּׁית",
    "שישי": "שִׁשִּׁי",
    "שישיות": "שִׁשִּׁיוֹת",
    "שישיים": "שִׁשִּׁיִּים",
    "שבע": "שֶׁ֫בַע",
    "שבעה": "שִׁבְעָה",
    "שבעת": "שִׁבְעַת",
    "שביעית": "שְׁבִיעִית",
    "שביעי": "שְׁבִיעִי",
    "שביעיות": "שְׁבִיעִיוֹת",
    "שביעיים": "שְׁבִיעִיִּים",
    "שמונה": "שְׁמ֫וֹנֶה",
    "שמונת": "שְׁמוֹנַת",
    "שמינית": "שְׁמִינִית",
    "שמיני": "שְׁמִינִי",
    "שמיניות": "שְׁמִינִיוֹת",
    "שמיניים": "שְׁמִינִיִּים",
    "תשע": "תֵּשַׁע",
    "תשעה": "תִּשְׁעָה",
    "תשעת": "תִּשְׁעַת",
    "תשיעית": "תְּשִׁיעִית",
    "תשיעי": "תְּשִׁיעִי",
    "תשיעיות": "תְּשִׁיעִיּוֹת",
    "תשיעיים": "תְּשִׁיעִיִּים",
}


TENS = {
    "עשר": "עֶ֫שֶׂר",
    "עשרה": "עֶשְׂרֵה",
    "עשרת": "עֲשֶׂ֫רֶת",
    "עשירית": "עֲשִׂירִית",
    "עשירי": "עֲשִׂירִי",
    "עשיריות": "עֲשִׂירִיּוֹת",
    "עשיריים": "עֲשִׂירִיִּים",
    "שתים עשרה": "שְׁתֵּ֫ים עֶשְׂרֵה",
    "שנים עשר": "שְׁנֵים עָשָׂר",
}


TWENTIES = {
    "עשרים": "עֶשְׂרִ֫ים",
    "שלושים": "שְׁלוֹשִׁים",
    "ארבעים": "אַרְבָּעִים",
    "חמישים": "חֲמִשִּׁים",
    "שישים": "שִׁשִּׁים",
    "שבעים": "שִׁבְעִים",
    "שמונים": "שְׁמוֹנִים",
    "תשעים": "תִּשְׁעִים",
}


HUNDREDS = {
    "מאה": "מֵ֫אָה",
    "מאת": "מֵאַת",
    "מאתיים": "מָאתַ֫יִם",
    "מאות": "מֵאוֹת",
}

THOUSANDS = {
    "אלף": "אֶ֫לֶף",
    "אלפיים": "אַלְפַּ֫יִם",
    "אלפים": "אֲלָפִים",
    "אלפי": "אַלְפִּי",
}


LARGE = {
    "מיליון": "מִילְיוֹן",
    "מיליוני": "מִילְיוֹנִי",
    "מיליארד": "מִילְיַארְד",
    "מיליארדי": "מִילְיַ֫ארְדִּי",
    "טריליון": "טְרִילְיוֹן",
    "טריליוני": "טְרִילְיוֹנִי",
    "קוודריליון": "קוֹוַדְרִילְיוֹן",
    "קוודריליוני": "קוֹוַדְרִילְיוֹנִי",
    "קווינטיליון": "קוִוִּנְטִילְיוֹן",
    "קווינטיליוני": "קוִוִּנְטִילְיוֹנִי",
    "סקסטיליון": "סְקֶסְטִילְיוֹן",
    "סקסטיליוני": "סְקֶסְטִילְיוֹנִי",
    "ספטיליון": "סְפֶּטִילְיוֹן",
    "ספטיליוני": "סְפֶּטִילְיוֹנִי",
    "אוקטיליון": "אוֹקְטִילְיוֹן",
    "אוקטיליוני": "אוֹקְטִילְיוֹנִי",
    "נוניליון": "נוּנִילְיוֹן",
    "נוניליוני": "נוּנִילְיוֹנִי",
    "דסיליון": "דֶּסִילְיוֹן",
    "דסיליוני": "דֶּסִילְיוֹנִי",
    "אונדסיליון": "אוּנְדְסִילְיוֹן",
    "אונדסיליוני": "אוּנְדְסִילְיוֹנִי",
    "דואודסיליון": "דוּאודְסִילְיוֹן",
    "דואודסיליוני": "דוּאודְסִילְיוֹנִי",
    "טרדסיליון": "טֶרְדְסִילְיוֹן",
    "טרדסיליוני": "טֶרְדְסִילְיוֹנִי",
    "קווטואורדסיליון": "קוּוטְוָאורְדְסִילְיוֹן",
    "קווטואורדסיליוני": "קוּוטְוָאורְדְסִילְיוֹנִי",
    "קווינדסיליון": "קוִוִּנְדְסִילְיוֹן",
    "קווינדסיליוני": "קוִוִּנְדְסִילְיוֹנִי",
    "סקסדסיליון": "סֶקְסְדְסִילְיוֹן",
    "סקסדסיליוני": "סֶקְסְדְסִילְיוֹנִי",
    "ספטנדסיליון": "סְפֶּטַנְדְסִילְיוֹן",
    "ספטנדסיליוני": "סְפֶּטַנְדְסִילְיוֹנִי",
    "אוקטודסיליון": "אוֹקְטוֹדְסִילְיוֹן",
    "אוקטודסיליוני": "אוֹקְטוֹדְסִילְיוֹנִי",
    "נובמדסיליון": "נוֹבְמַדְסִילְיוֹן",
    "נובמדסיליוני": "נוֹבְמַדְסִילְיוֹנִי",
    "ויגינטיליון": "וִיגִּינְטִילְיוֹן",
    "ויגינטיליוני": "וִיגִּינְטִילְיוֹנִי",
}


LETTERS = {
    "ו": "וֵ",
    "ה": "הַ",
}


CURRENCY = {
    "שקל": "שֵׁ֫קֶל",
    "שקלים": "שְׁקָלִים",
    "אגורה": "אֲגוֹרָה",
    "אגורות": "אֲגוֹרוֹת",
    "אירו": "אֵ֫ירוֹ",
    "סנט": "סֵנְט",
    "סנטים": "סֵ֫נְטִים",
    "דולר": "ד֫וֹלָר",
    "דולרים": "דוֹלָ֫רִים",
}


POINTS = {
    "מינוס": "מִ֫ינּוּס",
    "נקודה": "נְֽקֻדָּה",
}

NUMBER_NAMES = {
    **CURRENCY,
    **HUNDREDS,
    **LARGE,
    **LETTERS,
    **ONES,
    **POINTS,
    **TENS,
    **THOUSANDS,
    **TWENTIES,
    **ZERO,
}



"""
wget https://huggingface.co/thewh1teagle/phonikud-onnx/resolve/main/phonikud-1.0.int8.onnx
uv sync
uv run python examples/simple.py
"""

from phonikud_onnx import Phonikud
from phonikud import lexicon


def main():
    phonikud = Phonikud("./phonikud-1.0.int8.onnx")
    sentence = "הילדים אהבו במיוחד את הסיפורים הללו שהמורה הקריאה."
    with_diacritics = phonikud.add_diacritics(
        sentence, mark_matres_lectionis=lexicon.NIKUD_HASER_DIACRITIC
    )
    print(with_diacritics)


if __name__ == "__main__":
    main()



"""
uv run pytest tests/test_syllables.py
"""

import pandas as pd
from phonikud.syllables import add_stress, sort_diacritics
from pathlib import Path

base_path = Path(__file__).parent / "syllables_test_tables"
files = base_path.glob("*.csv")

STRESS = "\u05ab"
DAGESH = "\u05bc"


def check_syllables(filename):
    df = pd.read_csv(filename)
    errors = []

    for row in df.itertuples():
        src_word = sort_diacritics(row.word)
        # src_word = src_word.replace(DAGESH, "")
        src_without_stress = src_word.replace(STRESS, "")
        dst_with_stress = add_stress(src_without_stress, int(row.stress_index))

        dst_with_stress = sort_diacritics(dst_with_stress)
        if src_word != dst_with_stress:
            errors.append(
                f"{src_word.encode('unicode_escape')} \nVS\n {dst_with_stress.encode('unicode_escape')}\n{src_word} != {dst_with_stress}"
            )
    assert not errors, "\n".join(errors) + f"\nFound {len(errors)} errors\n"


def test_syllables():
    for file in files:
        check_syllables(file)



from phonikud.expander import Expander
from phonikud.utils import remove_nikud


def test_numbers():
    expander = Expander()
    text = expander.expand_text("35")
    assert "שלושים" in remove_nikud(text)



"""
uv pip install gradio
uv run gradio examples/editor.py
"""

from phonikud import phonemize, lexicon
from phonikud.utils import remove_nikud
import gradio as gr
from phonikud_onnx import Phonikud
from pathlib import Path


default_text = """
הַדַּיָּיג נִצְמָד לְדֹופֶן הַסִּירָה בִּזְמַן הַסְּעָרָה.  
הִסְבַּרְתִּי לָהּ אֶת הַכֹּל, וְאָמַרְתִּי בְּדִיּוּק מָה קָרָה.
הַיְּלָדִים אָהֲבוּ בִּמְיֻוחָד אֶת הַסִּיפּוּרִים הַלָּלוּ שֶׁהַמּוֹרָה הִקְרִיאָה.
""".strip()


def on_phonikud_toggle(use_phonikud):
    if not use_phonikud:
        return default_text
    return remove_nikud(default_text)


css = """
    .input textarea {
        font-size: 22px;
        padding: 15px;
        height: 200px;
    }

    .phonemes {
        background: var(--input-background-fill);
        
    }
    .phonemes {
        padding: 5px;
        min-height: 50px;    
    }
"""

theme = gr.themes.Soft(font=[gr.themes.GoogleFont("Noto Sans Hebrew")])

phonikud = None
model_path = Path("./phonikud-1.0.int8.onnx")
if model_path.exists():
    phonikud = Phonikud(str(model_path))


def on_submit(text: str, schema: str, use_phonikud: bool) -> str:
    diacritized = (
        phonikud.add_diacritics(
            text, mark_matres_lectionis=lexicon.NIKUD_HASER_DIACRITIC
        )
        if phonikud and use_phonikud
        else text
    )
    phonemes = phonemize(
        diacritized, predict_stress=True, schema=schema, predict_shva_nah=False
    )
    if use_phonikud:
        return f"<div dir='rtl' style='font-size: 22px;'>{diacritized.strip()}</div><br><div dir='ltr' style='font-size: 22px;'>{phonemes.strip()}</div>"
    else:
        return f"<div dir='ltr' style='font-size: 22px;'>{phonemes.strip()}</div>"


with gr.Blocks(theme=theme, css=css) as demo:
    text_input = gr.Textbox(
        value=remove_nikud(default_text),
        label="Text",
        rtl=True,
        elem_classes=["input"],
        lines=7,
    )

    with gr.Row():
        schema_dropdown = gr.Dropdown(
            choices=["modern", "plain"], value="plain", label="Phoneme Schema"
        )
        use_phonikud_checkbox = gr.Checkbox(
            value=True, label="Use Phonikud (add diacritics)"
        )

    submit_button = gr.Button("Create")
    output_box = gr.Markdown(label="Phonemes + Diacritics", elem_classes=["phonemes"])
    use_phonikud_checkbox.change(
        fn=lambda use_phonikud: (
            on_phonikud_toggle(use_phonikud),  # Update text_input
            on_submit(
                on_phonikud_toggle(use_phonikud), schema_dropdown.value, use_phonikud
            ),  # Update output_box
        ),
        inputs=use_phonikud_checkbox,
        outputs=[text_input, output_box],  # Update both text input and output box
    )

    submit_button.click(
        fn=on_submit,
        inputs=[text_input, schema_dropdown, use_phonikud_checkbox],
        outputs=output_box,
    )

    gr.Markdown("""
        <p style='text-align: center;'><a href='https://github.com/thewh1teagle/phonikud' target='_blank'>Phonikud on Github</a></p>
    """)

if __name__ == "__main__":
    demo.launch()



from functools import lru_cache
from typing import Literal
from phonikud import lexicon
import unicodedata
import regex as re
import phonikud.syllables
from phonikud.variants import Letter
import phonikud


def sort_diacritics(match):
    letter = match.group(1)
    diacritics = "".join(sorted(match.group(2)))  # Sort diacritics
    return letter + diacritics


NORMALIZE_PATTERNS = {
    # Sort diacritics
    r"(\p{L})(\p{M}+)": sort_diacritics,
    "״": '"',  # Hebrew geresh to normal geresh
    "׳": "'",  # Same
}


def remove_nikud(text: str, to_keep=""):
    pattern = lexicon.HE_NIKUD_PATTERN
    pattern = "".join(i for i in pattern if i not in to_keep)
    return re.sub(
        pattern,
        "",
        text,
    )


@lru_cache(maxsize=10000)
def normalize(text: str) -> str:
    """
    Normalize unicode (decomposite)
    Keep only Hebrew characters / punctuation / IPA
    Sort diacritics
    """

    # Decompose text
    text = unicodedata.normalize("NFD", text)
    for k, v in NORMALIZE_PATTERNS.items():
        text = re.sub(k, v, text)
    for k, v in lexicon.DEDUPLICATE.items():
        text = re.sub(k, v, text)
    return text


def post_normalize(phonemes: str):
    new_phonemes = []
    for word in phonemes.split(" "):
        # remove glottal stop from end
        word = re.sub(r"ʔ$", "", word)
        # remove h from end
        word = re.sub(r"h$", "", word)
        word = re.sub(r"ˈh$", "", word)
        # remove j followed by a i
        word = re.sub(r"ij$", "i", word)
        new_phonemes.append(word)
    phonemes = " ".join(new_phonemes)
    return phonemes


def post_clean(phonemes: str):
    clean = []
    for i in phonemes:
        if i == "-":
            clean.append(" ")
        elif (
            i in lexicon.SET_PHONEMES
            or i in lexicon.ADDITIONAL_PHONEMES
            or i == " "
            or i in lexicon.PUNCTUATION
        ):
            clean.append(i)
    return "".join(clean)


letters_pattern = re.compile(r"(\p{L})([\p{M}'|]*)")


# @lru_cache(maxsize=10000) TODO?
def get_letters(word: str):
    letters: list[tuple[str, str]] = letters_pattern.findall(word)  # with en_geresh
    letters: list[Letter] = [Letter(i[0], i[1]) for i in letters]
    return letters


def get_unicode_names(text: str):
    return [unicodedata.name(c, "?") for c in text]


def has_vowel(s: iter):
    return any(i in s for i in "aeiou")


def has_constant(s: iter):
    return any(i not in "aeiou" for i in s)


def get_phoneme_syllables(phonemes: list[str]) -> list[str]:
    syllables = []
    cur_syllable = ""

    i = 0
    while i < len(phonemes):
        # Add current phoneme to the syllable

        cur_syllable += phonemes[i]

        # If we have a vowel in the current syllable
        if has_vowel(cur_syllable):
            # If there's a next phoneme that's a consonant followed by a vowel-containing phoneme
            if (
                i + 2 < len(phonemes)
                and not has_vowel(phonemes[i + 1])
                and has_vowel(phonemes[i + 2])
            ):
                # End the current syllable and start a new one
                syllables.append(cur_syllable)
                cur_syllable = ""
            # If we're at the end or next phoneme has a vowel
            elif i + 1 >= len(phonemes) or has_vowel(phonemes[i + 1]):
                # End the current syllable
                syllables.append(cur_syllable)
                cur_syllable = ""

        i += 1

    # Add any remaining syllable
    if cur_syllable:
        syllables.append(cur_syllable)

    # Iterate over syllables and move any syllable ending with lexicon.STRESS to the next one
    for i in range(len(syllables) - 1):  # Ensure we're not at the last syllable
        if syllables[i].endswith(lexicon.STRESS_PHONEME):
            syllables[i + 1] = (
                lexicon.STRESS_PHONEME + syllables[i + 1]
            )  # Move stress to next syllable
            syllables[i] = syllables[i][
                : -len(lexicon.STRESS_PHONEME)
            ]  # Remove stress from current syllable

    return syllables


def sort_stress(
    phonemes: list[str], placement: Literal["syllable", "vowel"] = "vowel"
) -> list[str]:
    """
    TTS systems expect that the stress will be BEFORE vowel
    Linguistics expect in the START of the syllable
    at_start=True for place it in the beginning
    """
    if "ˈ" not in "".join(phonemes):
        # ^ Does not contains stress
        return phonemes
    if not any(i in "".join(phonemes) for i in "aeiou"):
        # ^ Does not contains vowel
        return phonemes

    # Remove stress marker
    phonemes = [p for p in phonemes if p != "ˈ"]

    if placement == "syllable":
        return ["ˈ"] + phonemes

    # Define vowels
    vowels = "aeiou"

    # Find the first phoneme that contains a vowel, and inject the stress before the vowel

    for i, phoneme in enumerate(phonemes):
        for j, char in enumerate(phoneme):
            if char in vowels:
                # Insert stress before the vowel
                phonemes[i] = phoneme[:j] + "ˈ" + phoneme[j:]
                return phonemes

    # If no vowels found, return unchanged
    return phonemes


def mark_shva_na(word: str):
    """
    Shva Na is context-independent and can be predicted with just the word or a dictionary.
    See https://hebrew-academy.org.il/2020/08/11/איך-הוגים-את-השווא-הנע
    Note: we predict only if Shva in the first letter in the word
    Note: we assume that the word comes with | to mark 'Txiliyot'
    Note: Shva Na rules mid-word are unreliable, so we don’t code them.

    Meteg (\u05bd) will be added in the letter with Shva Na

    What we don't predict:
    (1) some shva in beginning in future form (we don't know)
    (2) shva in the middle of the word
    """
    letters = get_letters(word)
    if not letters:
        return word
    if letters[0].char in "למנרי":
        letters[0].all_diac += lexicon.SHVA_NA_DIACRITIC
    elif len(letters) > 1 and letters[1].char in "אעה":
        letters[0].all_diac += lexicon.SHVA_NA_DIACRITIC
    elif letters[0].char in "וכלב" and lexicon.PREFIX_DIACRITIC in letters[0].all_diac:
        # ^ The nakdan should add |
        letters[0].all_diac += lexicon.SHVA_NA_DIACRITIC
    # Ensure that prefix character will be last
    for letter in letters:
        if "|" in letter.all_diac:
            letter.all_diac = letter.all_diac.replace("|", "") + "|"
    return "".join(str(i) for i in letters)


def sort_hatama(letters: list[Letter]) -> list[Letter]:
    for i in range(len(letters) - 1):
        diacs = list(letters[i].all_diac)
        if lexicon.HATAMA_DIACRITIC in diacs and lexicon.NIKUD_HASER_DIACRITIC in diacs:
            diacs.remove(lexicon.HATAMA_DIACRITIC)
            letters[i].all_diac = "".join(diacs)  # Reassign the updated diacritics
            letters[i + 1].all_diac += lexicon.HATAMA_DIACRITIC
    return letters


def add_milra_hatama(word: str):
    syllables = phonikud.syllables.get_syllables(word)
    stress_index = -1

    if not syllables:
        return word

    if len(syllables) == 1:
        stress_index = 0

    # Get latest syllable
    milra = syllables[stress_index]
    # Get letters
    letters = get_letters(milra)
    # Add Hatama
    letters[0].all_diac += lexicon.HATAMA_DIACRITIC

    # Replace latest syllable
    syllables[stress_index] = "".join(str(i) for i in letters)
    return "".join(syllables)



from typing import List
import torch
from src.train.config import TrainArgs
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data import DataLoader, Dataset
from transformers import BertTokenizerFast
from typing import Tuple
from src.model.phonikud_model import (
    HATAMA_CHAR,
    MOBILE_SHVA_CHAR,
    PREFIX_CHAR,
)


def get_dataloader(lines: list[str], args: TrainArgs, collator: "Collator"):
    train_data = TrainData(lines)

    loader = DataLoader(
        train_data,
        batch_size=args.batch_size,
        shuffle=True,
        collate_fn=collator.collate_fn,
        num_workers=args.num_workers,
    )
    return loader


class TrainData(Dataset):
    def __init__(self, lines: List[str]):
        self.lines = lines

    def __len__(self):
        return len(self.lines)

    def __getitem__(self, idx):
        line = self.lines[idx]
        text = ""
        num_classes = 4
        targets = []  # list of [num_classes] binary arrays

        for char in line:
            # Feature char (apply to previous real char)
            if char in (HATAMA_CHAR, MOBILE_SHVA_CHAR, PREFIX_CHAR):
                # Map feature char to class index and queue it
                label_idx = {HATAMA_CHAR: 1, MOBILE_SHVA_CHAR: 2, PREFIX_CHAR: 3}[char]
                if len(targets) > 0:
                    # Set the label for the previous char (multi-label allowed)
                    targets[-1][label_idx] = 1
            else:
                # Real Hebrew character: start with no labels
                arr = [0] * num_classes
                arr[0] = 1  # Default "plain" unless any other is present
                targets.append(arr)
                text += char

        # Remove "plain" (arr[0]) when any other label is set (optional)
        for arr in targets:
            if sum(arr[1:]) > 0:
                arr[0] = 0

        assert len(text) == len(targets)

        # Output shape: [seq_len, num_classes], dtype float
        targets_tensor = torch.tensor(targets, dtype=torch.float)
        return text, targets_tensor


class Collator:
    def __init__(self, tokenizer: BertTokenizerFast):
        self.tokenizer = tokenizer

    def collate_fn(self, items: List[Tuple[str, torch.Tensor]]):
        inputs = self.tokenizer(
            [x[0] for x in items], padding=True, truncation=True, return_tensors="pt"
        )
        targets = pad_sequence([x[1] for x in items], batch_first=True)
        # ^ shape: (batch_size, n_chars_padded, n_active_components)

        return inputs, targets



import time
from phonikud import phonemize

input_text = "אִישׁ לֹא טָרַח בְּֽמֶשֶׁךְ יוֹתֵר מֵחֹ֫ודֶשׁ לְֽהָכִין אֶת הַמַּעֲרָכוֹת הַשּׁוֹנוֹת לַקָּטַסְטְרוֹפָה הַהֲמוֹנִית שֶׁעֲלוּלָה לְֽהִתְרַחֵשׁ כָּאן."

start_time = time.time()
for _ in range(20):
    phonemized_text = phonemize(input_text)
end_time = time.time()

# Calculate results
elapsed_time = end_time - start_time
avg_time = elapsed_time / 20

print(f"⏱️ Total: {elapsed_time:.6f}s | Avg: {avg_time:.6f}s per sentence")



"""
pip install -U phonikud-onnx phonikud

wget https://huggingface.co/thewh1teagle/phonikud/resolve/main/phonikud-1.0.int8.onnx
uv run examples/with_phonikud_onnx.py
"""

from phonikud_onnx import Phonikud
from phonikud import phonemize

phonikud = Phonikud("./phonikud-1.0.int8.onnx")
sentence = "בשנת 1948 השלים אפרים קישון את לימודיו בפיסול מתכת ובתולדות האמנות והחל לפרסם מאמרים הומוריסטיים על כל מה שרצה"
with_diacritics = phonikud.add_diacritics(sentence)
phonemes = phonemize(with_diacritics)
print("Sentence: ", with_diacritics)
print("Phonemes: ", phonemes)



import torch
from torch import nn
from dataclasses import dataclass
from transformers.utils import ModelOutput
from transformers import BertTokenizerFast
import re


from .dicta_model import (
    BertForDiacritization,
    is_hebrew_letter,
    is_matres_letter,
)

HATAMA_CHAR = "\u05ab"  # "ole" symbol marks hatama
MOBILE_SHVA_CHAR = "\u05bd"  # "meteg" symbol marks shva na (mobile shva)
PREFIX_CHAR = "|"  # vertical bar
NIKUD_HASER = "\u05af"  # not in use but dicta has it

PHONETIC_NIKUD = HATAMA_CHAR + MOBILE_SHVA_CHAR + PREFIX_CHAR + NIKUD_HASER


def remove_nikud(text: str, additional=""):
    """
    Remove nikud except meteg as we use it for Shva Na
    """
    return re.sub(f"[\u05b0-\u05bc\u05be-\u05c7{additional}]", "", text)


@dataclass
class MenakedLogitsOutput(ModelOutput):
    nikud_logits: torch.FloatTensor = None
    shin_logits: torch.FloatTensor = None
    additional_logits: torch.FloatTensor = None  # For hatama, mobile shva, and prefix

    def detach(self):
        return MenakedLogitsOutput(
            self.nikud_logits.detach(),
            self.shin_logits.detach(),
            self.additional_logits.detach(),
        )


class PhoNikudModel(BertForDiacritization):
    def __init__(self, config):
        super().__init__(config)
        self.config = config
        self.mlp = nn.Sequential(nn.Linear(1024, 100), nn.ReLU(), nn.Linear(100, 4))
        # ^ predicts hatama, mobile shva, and prefix; outputs are logits

    def freeze_base_model(self):
        self.bert.eval()
        self.menaked.eval()
        for param in self.bert.parameters():
            param.requires_grad = False

        for param in self.menaked.parameters():
            param.requires_grad = False

    def forward(self, x):
        # based on: https://huggingface.co/dicta-il/dictabert-large-char-menaked/blob/main/BertForDiacritization.py
        bert_outputs = self.bert(**x)
        hidden_states = bert_outputs.last_hidden_state
        # ^ shape: (batch_size, n_chars_padded, 1024)
        hidden_states = self.dropout(hidden_states)

        _, nikud_logits = self.menaked(hidden_states)
        # ^ nikud_logits: MenakedLogitsOutput

        additional_logits = self.mlp(hidden_states)
        # ^ shape: (batch_size, n_chars_padded, 3) [hatama, mobile shva, and prefix]

        return MenakedLogitsOutput(
            nikud_logits.nikud_logits, nikud_logits.shin_logits, additional_logits
        )

    def encode(
        self, sentences: list[str], tokenizer: BertTokenizerFast, padding="longest"
    ):
        sentences = [remove_nikud(sentence) for sentence in sentences]

        # Assert the lengths are within the tokenizer's max limit
        assert all(
            len(sentence) + 2 <= tokenizer.model_max_length for sentence in sentences
        ), f"All sentences must be <= {tokenizer.model_max_length}, please segment and try again"

        # Tokenize the inputs and return the tensor format
        inputs = tokenizer(
            sentences,
            padding=padding,
            truncation=True,
            return_tensors="pt",
            return_offsets_mapping=True,
        )
        offset_mapping = inputs.pop("offset_mapping")
        inputs = {k: v.to(self.device) for k, v in inputs.items()}

        return inputs, offset_mapping

    def decode(
        self,
        sentences,
        offset_mapping,
        nikud_predictions,
        shin_predictions,
        hatama_predictions,
        mobile_shva_predictions,
        prefix_predictions,
        mark_matres_lectionis: str = None,
    ):
        ret = []
        for sent_idx, (sentence, sent_offsets) in enumerate(
            zip(sentences, offset_mapping)
        ):
            output = []
            prev_index = 0
            for idx, offsets in enumerate(sent_offsets):
                # Add anything missed
                if offsets[0] > prev_index:
                    output.append(sentence[prev_index : offsets[0]])
                if offsets[1] - offsets[0] != 1:
                    continue

                # Get the next character
                char = sentence[offsets[0] : offsets[1]]
                prev_index = offsets[1]

                if not is_hebrew_letter(char):
                    output.append(char)
                    continue

                # Apply the predictions to the character
                nikud = self.config.nikud_classes[nikud_predictions[sent_idx][idx]]
                shin = (
                    ""
                    if char != "ש"
                    else self.config.shin_classes[shin_predictions[sent_idx][idx]]
                )

                # Check for matres lectionis
                if nikud == self.config.mat_lect_token:
                    if not is_matres_letter(char):
                        nikud = ""  # Don't allow matres on irrelevant letters
                    elif mark_matres_lectionis is not None:
                        nikud = mark_matres_lectionis
                    else:
                        continue

                # Apply hatama, mobile shva, and prefix predictions
                hatama = HATAMA_CHAR if hatama_predictions[sent_idx][idx] == 1 else ""
                mobile_shva = (
                    MOBILE_SHVA_CHAR
                    if mobile_shva_predictions[sent_idx][idx] == 1
                    else ""
                )
                prefix = PREFIX_CHAR if prefix_predictions[sent_idx][idx] == 1 else ""

                output.append(char + shin + nikud + hatama + mobile_shva + prefix)

            output.append(sentence[prev_index:])
            ret.append("".join(output))

        return ret

    def create_predictions(self, inputs):
        output = self.forward(inputs)

        # Extract the logits from the model output
        nikud_logits = output.detach()
        additional_logits = output.additional_logits.detach()
        # additional_logits = additional_logits.permute(0, 2, 1)

        # Get predictions from logits
        nikud_predictions = nikud_logits.nikud_logits.argmax(dim=-1).tolist()
        shin_predictions = nikud_logits.shin_logits.argmax(dim=-1).tolist()

        hatama_predictions = (additional_logits[..., 1] > 1).int().tolist()
        mobile_shva_predictions = (additional_logits[..., 2] > 1).int().tolist()
        prefix_predictions = (additional_logits[..., 3] > 1).int().tolist()

        return (
            nikud_predictions,
            shin_predictions,
            hatama_predictions,
            mobile_shva_predictions,
            prefix_predictions,
        )

    @torch.no_grad()
    def predict(
        self,
        sentences,
        tokenizer: BertTokenizerFast,
        mark_matres_lectionis: str = None,
        padding="longest",
    ):
        # Step 1: Encoding (tokenizing sentences)
        inputs, offset_mapping = self.encode(sentences, tokenizer, padding)

        # Step 2: Making predictions
        (
            nikud_predictions,
            shin_predictions,
            hatama_predictions,
            mobile_shva_predictions,
            prefix_predictions,
        ) = self.create_predictions(inputs)

        # Step 3: Decoding (reconstructing the sentences with predictions)
        result = self.decode(
            sentences,
            offset_mapping,
            nikud_predictions,
            shin_predictions,
            hatama_predictions,
            mobile_shva_predictions,
            prefix_predictions,
            mark_matres_lectionis,
        )

        return result



from phonikud.utils import mark_shva_na, normalize
import pandas as pd
from phonikud import lexicon
from pathlib import Path

BASE_PATH = Path(__file__).parent


def test_shva_na():
    df = pd.read_csv(BASE_PATH / "./shva_test_tables/words.csv")
    for row in df.itertuples():
        src_word = row.word
        dst_word = src_word.replace(lexicon.SHVA_NA_DIACRITIC, "")
        dst_word = mark_shva_na(dst_word)

        # Normalize
        src_word = normalize(src_word)
        dst_word = normalize(dst_word)

        assert dst_word == src_word



"""
Convert time to words
TODO: fix zeros eg. 22:00
"""

import re

PATTERNS = [
    r"(\d{1,2})([apm]{2})",  # AM/PM format
    r"(\d{1,2}):(\d{2})",  # HH:MM format
]


def extract_time(match):
    """
    Extract hour and minute from a string in HH:MM or AM/PM format
    and return as integers.
    """
    time_str = match.group(0).lower().strip()

    # Check for HH:MM format
    match = re.match(r"(\d{1,2}):(\d{2})", time_str)
    if match:
        h = int(match.group(1))
        m = int(match.group(2))
        return f"{convert_to_word(h, m)}"

    # Check for AM/PM format
    match = re.match(r"(\d{1,2})([apm]{2})", time_str)
    if match:
        h = int(match.group(1))
        period = match.group(2)

        # Normalize to 24-hour format
        if period == "am" and h == 12:
            h = 0
        elif period == "pm" and h != 12:
            h += 12
        return f"{convert_to_word(h, 0)}"  # Defaulting to 0 minutes when only hour is provided

    return match.group(0)  # Return original text if the format is not recognized


def convert_to_word(h, m):
    hours = [
        "אֶפֶס",
        "אַחַת",
        "שְׁנַיִם",  # Will be replaced with "שֵׁנִי" when needed
        "שָׁלוֹשׁ",
        "אַ֫רְבַּע",
        "חָמֵשׁ",
        "שֵׁשׁ",
        "שֶׁ֫בַע",
        "שְׁמ֫וֹנֵה",
        "תֵּ֫שַׁע",
        "עֵ֫שֵׂר",
        "אַחַת עֶשְׂרֵה",
        "שְׁתֵּים עֶשְׂרֵה",
    ]

    tens = ["", "עֵשֵׂר", "עֶשְׂרִים", "שְׁלוֹשִׁים", "אַרְבָּעִים", "חֲמִשִּׁים"]

    ten_to_twenty = [
        "עֵ֫שֵׂר",
        "אַחַת עֶשְׂרֵה",
        "שְׁתֵּים עֶשְׂרֵה",
        "שְׁלוֹשׁ עֶשְׂרֵה",
        "אַרְבַּע עֶשְׂרֵה",
        "חֲמֵשׁ עֶשְׂרֵה",
        "שֵׁשׁ עֶשְׂרֵה",
        "שְׁבַע עֶשְׂרֵה",
        "שְׁמוֹנֶה עֶשְׂרֵה",
        "תְּשַׁע עֶשְׂרֵה",
    ]

    vocab = {"minutes": "דַּקּוֹת", "and": "וֵ", "shtey": "שְׁתֵּי"}

    # Convert 0 hours to 12 (midnight)
    if h == 0:
        h = 12

    elif h > 12:
        h -= 12

    if m == 0:
        return f"{hours[h]}"

    elif 1 <= m <= 9:
        minute_word = (
            vocab["shtey"] if m == 2 else hours[m]
        )  # Replace "שניים" with "שני"
        return f"{hours[h]} {vocab['and']}{minute_word} {vocab['minutes']}"

    elif 10 <= m <= 19:
        return f"{hours[h]} {vocab['and']}{ten_to_twenty[m - 10]} {vocab['minutes']}"

    else:
        tens_part = f"{vocab['and']}{tens[m // 10]}"
        units_part = f"{vocab['and']}{hours[m % 10]}" if m % 10 != 0 else ""
        return f"{hours[h]} {tens_part} {units_part} {vocab['minutes']}".strip()


def time_to_word(text: str):
    return re.sub("|".join(PATTERNS), extract_time, text)



"""
ASCII IPA transcription of Hebrew consonants and vowels.
"""

# https://en.wikipedia.org/wiki/Unicode_and_HTML_for_the_Hebrew_alphabet#Compact_table

# Non standard diacritics
SHVA_NA_DIACRITIC = "\u05bd"  # Meteg
HATAMA_DIACRITIC = "\u05ab"  # Ole
PREFIX_DIACRITIC = "|"  # Vertical bar
NIKUD_HASER_DIACRITIC = "\u05af"  # Masora, not in use
EN_GERESH = "'"
NON_STANDARD_DIAC = "".join(
    [
        SHVA_NA_DIACRITIC,
        HATAMA_DIACRITIC,
        PREFIX_DIACRITIC,
        NIKUD_HASER_DIACRITIC,
        EN_GERESH,
    ]
)

HE_PATTERN = rf'[\u05b0-\u05ea{NON_STANDARD_DIAC}"]+'
# ^ Standard nikud and letters, ole, meteg, masora, vertical bar, en geresh
HE_NIKUD_PATTERN = rf"[\u05b0-\u05c7{NON_STANDARD_DIAC}]"
# ^ Letters, diacritics, en geresh
PUNCTUATION = set(r".,!? ")

STRESS_PHONEME = "ˈ"  # \u02c8 visually looks like single quote
SPECIAL_PHONEMES = ["w"]
MODERN_SCHEMA = {
    "x": "χ",  # Het
    "r": "ʁ",  # Resh
    "g": "ɡ",  # Gimel
}

# Geresh
GERESH_PHONEMES = {"ג": "dʒ", "ז": "ʒ", "ת": "ta", "צ": "tʃ", "ץ": "tʃ"}

# Consonants
LETTERS_PHONEMES = {
    "א": "ʔ",  # Alef
    "ב": "v",  # Bet
    "ג": "g",  # Gimel
    "ד": "d",  # Dalet
    "ה": "h",  # He
    "ו": "v",  # Vav
    "ז": "z",  # Zayin
    "ח": "x",  # Het
    "ט": "t",  # Tet
    "י": "j",  # Yod
    "ך": "x",  # Haf sofit
    "כ": "x",  # Haf
    "ל": "l",  # Lamed
    "ם": "m",  # Mem Sofit
    "מ": "m",  # Mem
    "ן": "n",  # Nun Sofit
    "נ": "n",  # Nun
    "ס": "s",  # Samekh
    "ע": "ʔ",  # Ayin, only voweled
    "פ": "f",  # Fey
    "ף": "f",  # Fey Sofit
    "ץ": "ts",  # Tsadik sofit
    "צ": "ts",  # Tsadik
    "ק": "k",  # Kuf
    "ר": "r",  # Resh
    "ש": "ʃ",  # Shin
    "ת": "t",  # Taf
    # Beged Kefet
    "בּ": "b",
    "כּ": "k",
    "פּ": "p",
    # Shin Sin
    "שׁ": "ʃ",
    "שׂ": "s",
    "'": "",
}

NIKUD_PHONEMES = {
    "\u05b4": "i",  # Hiriq
    "\u05b1": "e",  # Hataf segol
    "\u05b5": "e",  # Tsere
    "\u05b6": "e",  # Segol
    "\u05b2": "a",  # Hataf Patah
    "\u05b7": "a",  # Patah
    "\u05c7": "o",  # Kamatz katan
    "\u05b9": "o",  # Holam
    "\u05ba": "o",  # Holam haser for vav
    "\u05bb": "u",  # Qubuts
    "\u05b3": "o",  # Hataf qamats
    "\u05b8": "a",  # Kamataz
    HATAMA_DIACRITIC: STRESS_PHONEME,  # Stress (Hat'ama)
    SHVA_NA_DIACRITIC: "e",  # Shva na
}

DEDUPLICATE = {
    "\u05f3": "'",  # Hebrew geresh to regular geresh
    "־": "-",  # Hebrew Makaf to hypen
}

# Sets
SET_PHONETIC_DIACRITICS = set([HATAMA_DIACRITIC, PREFIX_DIACRITIC, SHVA_NA_DIACRITIC])

ADDITIONAL_PHONEMES = set()  # When using fallback
SET_PHONEMES = set(
    sorted(
        {
            *NIKUD_PHONEMES.values(),
            *LETTERS_PHONEMES.values(),
            *GERESH_PHONEMES.values(),
            *MODERN_SCHEMA.values(),
            *SPECIAL_PHONEMES,
        }
    )
)



from tap import Tap
from typing import Literal
from pathlib import Path

BASE_PATH = Path(__file__).parent / "../.."


class TrainArgs(Tap):
    model_checkpoint: str = "dicta-il/dictabert-large-char-menaked"
    "Path or name of the pretrained model checkpoint"

    device: Literal["cuda", "cpu", "mps"] = "cuda"

    data_dir: str = BASE_PATH / "data/train"
    "Path with txt files for train"

    output_dir: str = BASE_PATH / "ckpt/"
    "Path to save checkpoints"

    batch_size: int = 16
    "Batch size"

    epochs: int = 1000
    "Train epochs"

    pre_training_step: int = 0
    "Resume training from this step to preserve checkpoint name"

    learning_rate: float = 1e-3
    "Learning rate"

    early_stopping_patience = 10
    "Early stop if no improvement multiple times in checkpoint interval"

    num_workers: int = 0
    "Number of workers for data loading"

    checkpoint_interval: int = 100
    "Number of steps between saving checkpoints"

    val_split: float = 0.1
    "Fraction of training data to use as validation (0 to disable)"

    split_seed: int = 42
    "Random seed for train/val split"

    use_eval_file: bool = False
    "Use data/eval/*.txt as validation set instead of splitting train"


def get_opts():
    return TrainArgs().parse_args()



"""
Expand dates and numbers into words with nikud
This happens before phonemization
"""

from .numbers import num_to_word
from .dates import date_to_word
from .time_to_word import time_to_word
from .dictionary import Dictionary
from phonikud.log import log


class Expander:
    def __init__(self):
        self.dictionary = Dictionary()

    def expand_text(self, text: str):
        words = []
        for source_word in text.split():
            try:
                word = date_to_word(source_word)
                if word == source_word:
                    word = time_to_word(word)
                if word == source_word:
                    word = num_to_word(word)
                words.append(word)
            except Exception as e:
                log.error(f"Failed to expand {word} with error: {e}")
                words.append(source_word)
        text = " ".join(words)
        text = self.dictionary.expand_text(text)

        return text



import phonikud
from phonikud import lexicon


class Letter:
    def __init__(self, char: str, diac: list[str]):
        self.char = phonikud.normalize(char)
        self.all_diac = phonikud.normalize(diac)
        self.diac = "".join(
            i for i in self.all_diac if i not in lexicon.SET_PHONETIC_DIACRITICS
        )

    def __repr__(self):
        return f"[Letter] {self.char}{''.join(self.all_diac)}"

    def __eq__(self, value: "Letter"):
        return value.all_diac == self.all_diac and value.char == self.char

    def __str__(self):
        return self.char + self.all_diac



"""
Train from scratch:
    uv run src/train.py --device cpu --epochs 1
Train from checkpoint:
    uv run src/train.py --device cuda --epochs 5 --batch_size 8 --learning_rate 5e-4 --num_workers 2 \
        --model_checkpoint ./ckpt/step_21441_loss_0.0081/
    uv run src/train.py --device cuda --epochs 3 --model_checkpoint dicta-il/dictabert-large-char-menaked
On V100:
    uv run src/train.py --device cuda --epochs 5 --batch_size 64 --num_workers 8
    TODO: we may be able to train on very large epoch, with LR tunning
"""

from config import get_opts
from data import Collator, get_dataloader
from train_loop import train_model
from transformers import AutoTokenizer
from src.model.phonikud_model import PhoNikudModel
from torch.utils.tensorboard import SummaryWriter
from utils import print_model_size, read_lines


def main():
    args = get_opts()
    print(f"🧠 Loading model from {args.model_checkpoint}...")

    model = PhoNikudModel.from_pretrained(args.model_checkpoint, trust_remote_code=True)
    print_model_size(model)

    model.to(args.device)
    model.freeze_base_model()

    tokenizer = AutoTokenizer.from_pretrained(args.model_checkpoint, use_fast=True)
    collator = Collator(tokenizer)

    # Data split
    print("📖🔍 Reading lines from dataset...")
    train_lines, val_lines = read_lines(args.data_dir)
    print(
        f"✅ Loaded {len(train_lines)} training lines and {len(val_lines)} validation lines."
    )

    # Data loader
    train_dataloader = get_dataloader(train_lines, args, collator)
    val_dataloader = get_dataloader(val_lines, args, collator)

    # Log
    writer = SummaryWriter(log_dir=args.output_dir)

    # Train
    train_model(model, tokenizer, train_dataloader, val_dataloader, args, writer)


if __name__ == "__main__":
    main()
